```@meta
EditURL = "https://github.com/maartenvd/MPSKit.jl/examples/quantum1d/2.haldane/main.jl"
```

[![](https://mybinder.org/badge_logo.svg)](https://mybinder.org/v2/gh/maartenvd/MPSKit.jl/gh-pages?filepath=dev/examples/quantum1d/2.haldane/main.ipynb)
[![](https://img.shields.io/badge/show-nbviewer-579ACA.svg)](https://nbviewer.jupyter.org/github/maartenvd/MPSKit.jl/blob/gh-pages/dev/examples/quantum1d/2.haldane/main.ipynb)
[![](https://img.shields.io/badge/download-project-orange)](https://minhaskamal.github.io/DownGit/#/home?url=https://github.com/maartenvd/MPSKit.jl/examples/tree/gh-pages/dev/examples/quantum1d/2.haldane)

# [The Haldane gap](@id demo_haldane)

In this tutorial we will calculate the Haldane gap (the energy gap in the ``S = 1`` Heisenberg model) in two different ways.
To follow the tutorial you need the following packages:

````julia
using MPSKit, MPSKitModels, TensorKit, Plots, Polynomials
import TensorOperations; TensorOperations.disable_cache(); # hide
````

The Heisenberg model is defined by the following hamiltonian:

```math
H = -J∑_{⟨i,j⟩} (X_i X_j + Y_i Y_j + Z_i Z_j)
```

This hamiltonian has an SU(2) symmetry, which we can enforce by using SU(2)-symmetric tensors:

````julia
H = heisenberg_XXX(ComplexF64, SU2Irrep; spin=1, J=-1);
````

## Finite size extrapolation

We can start the analysis using finite-size methods.
The groundstate of this model can be approximated using finite MPS through the use of DMRG.

The typical way to find excited states is to minimize the energy while adding an error term
$$λ \left|gs\right> \left< gs\right|$$
Here we will instead use the [quasiparticle ansatz](https://journals.aps.org/prl/abstract/10.1103/PhysRevLett.111.080401).

In Steven White's original DMRG paper it was remarked that the ``S = 1`` excitations correspond to edge states, and that one should define the Haldane gap as the difference in energy between the ``S = 2`` and ``S = 1`` states.
This can be done as follows.

````julia
L = 10
physical_space = Rep[SU₂](1 => 1)
virtual_space = Rep[SU₂](0 => 15, 1 => 15, 2 => 10, 3 => 5)
ψ₀ = FiniteMPS(rand, ComplexF64, L, physical_space, virtual_space)
ψ, envs, delta = find_groundstate(ψ₀, H, DMRG(; verbose=false))
E₀ = real.(expectation_value(ψ, H))
En_1, st_1 = excitations(H, QuasiparticleAnsatz(), ψ, envs; sector=SU₂(1))
En_2, st_2 = excitations(H, QuasiparticleAnsatz(), ψ, envs; sector=SU₂(2))
ΔE_finite = real(En_2[1] - En_1[1])
````

````
-0.2897302384697755
````

We can go even further and doublecheck the claim that ``S = 1`` is an edge excitation, by plotting the energy density.

````julia
p_density = plot(; xaxis="position", yaxis="energy density")
excited_1 = convert(FiniteMPS, st_1[1])
excited_2 = convert(FiniteMPS, st_2[1])
plot!(p_density, real.(expectation_value(excited_1, H) .- E₀) ./ L; label="S = 1")
plot!(p_density, real.(expectation_value(excited_2, H) .- E₀) ./ L; label="S = 2")
````

```@raw html
<?xml version="1.0" encoding="utf-8"?>
<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" width="600" height="400" viewBox="0 0 2400 1600">
<defs>
  <clipPath id="clip340">
    <rect x="0" y="0" width="2400" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip340)" d="M0 1600 L2400 1600 L2400 0 L0 0  Z" fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip341">
    <rect x="480" y="0" width="1681" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip340)" d="M102.74 1505.26 L2352.76 1505.26 L2352.76 47.2441 L102.74 47.2441  Z" fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip342">
    <rect x="102" y="47" width="2251" height="1459"/>
  </clipPath>
</defs>
<polyline clip-path="url(#clip342)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="402.271,1505.26 402.271,47.2441 "/>
<polyline clip-path="url(#clip342)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="873.972,1505.26 873.972,47.2441 "/>
<polyline clip-path="url(#clip342)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="1345.67,1505.26 1345.67,47.2441 "/>
<polyline clip-path="url(#clip342)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="1817.37,1505.26 1817.37,47.2441 "/>
<polyline clip-path="url(#clip342)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="2289.08,1505.26 2289.08,47.2441 "/>
<polyline clip-path="url(#clip340)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="102.74,1505.26 2352.76,1505.26 "/>
<polyline clip-path="url(#clip340)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="402.271,1505.26 402.271,1486.36 "/>
<polyline clip-path="url(#clip340)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="873.972,1505.26 873.972,1486.36 "/>
<polyline clip-path="url(#clip340)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="1345.67,1505.26 1345.67,1486.36 "/>
<polyline clip-path="url(#clip340)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="1817.37,1505.26 1817.37,1486.36 "/>
<polyline clip-path="url(#clip340)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="2289.08,1505.26 2289.08,1486.36 "/>
<path clip-path="url(#clip340)" d="M396.923 1563.72 L413.243 1563.72 L413.243 1567.66 L391.298 1567.66 L391.298 1563.72 Q393.96 1560.97 398.544 1556.34 Q403.15 1551.69 404.331 1550.35 Q406.576 1547.82 407.456 1546.09 Q408.358 1544.33 408.358 1542.64 Q408.358 1539.88 406.414 1538.15 Q404.493 1536.41 401.391 1536.41 Q399.192 1536.41 396.738 1537.17 Q394.308 1537.94 391.53 1539.49 L391.53 1534.77 Q394.354 1533.63 396.808 1533.05 Q399.261 1532.47 401.298 1532.47 Q406.669 1532.47 409.863 1535.16 Q413.058 1537.85 413.058 1542.34 Q413.058 1544.47 412.247 1546.39 Q411.46 1548.28 409.354 1550.88 Q408.775 1551.55 405.673 1554.77 Q402.571 1557.96 396.923 1563.72 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M876.981 1537.17 L865.176 1555.62 L876.981 1555.62 L876.981 1537.17 M875.754 1533.1 L881.634 1533.1 L881.634 1555.62 L886.564 1555.62 L886.564 1559.51 L881.634 1559.51 L881.634 1567.66 L876.981 1567.66 L876.981 1559.51 L861.379 1559.51 L861.379 1555 L875.754 1533.1 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M1346.08 1548.52 Q1342.93 1548.52 1341.08 1550.67 Q1339.25 1552.82 1339.25 1556.57 Q1339.25 1560.3 1341.08 1562.47 Q1342.93 1564.63 1346.08 1564.63 Q1349.23 1564.63 1351.06 1562.47 Q1352.91 1560.3 1352.91 1556.57 Q1352.91 1552.82 1351.06 1550.67 Q1349.23 1548.52 1346.08 1548.52 M1355.36 1533.86 L1355.36 1538.12 Q1353.6 1537.29 1351.8 1536.85 Q1350.01 1536.41 1348.25 1536.41 Q1343.62 1536.41 1341.17 1539.53 Q1338.74 1542.66 1338.39 1548.98 Q1339.76 1546.97 1341.82 1545.9 Q1343.88 1544.81 1346.36 1544.81 Q1351.56 1544.81 1354.57 1547.98 Q1357.61 1551.13 1357.61 1556.57 Q1357.61 1561.9 1354.46 1565.11 Q1351.31 1568.33 1346.08 1568.33 Q1340.08 1568.33 1336.91 1563.75 Q1333.74 1559.14 1333.74 1550.41 Q1333.74 1542.22 1337.63 1537.36 Q1341.52 1532.47 1348.07 1532.47 Q1349.83 1532.47 1351.61 1532.82 Q1353.42 1533.17 1355.36 1533.86 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M1817.37 1551.25 Q1814.04 1551.25 1812.12 1553.03 Q1810.22 1554.81 1810.22 1557.94 Q1810.22 1561.06 1812.12 1562.85 Q1814.04 1564.63 1817.37 1564.63 Q1820.71 1564.63 1822.63 1562.85 Q1824.55 1561.04 1824.55 1557.94 Q1824.55 1554.81 1822.63 1553.03 Q1820.73 1551.25 1817.37 1551.25 M1812.7 1549.26 Q1809.69 1548.52 1808 1546.46 Q1806.33 1544.4 1806.33 1541.43 Q1806.33 1537.29 1809.27 1534.88 Q1812.24 1532.47 1817.37 1532.47 Q1822.54 1532.47 1825.48 1534.88 Q1828.42 1537.29 1828.42 1541.43 Q1828.42 1544.4 1826.73 1546.46 Q1825.06 1548.52 1822.07 1549.26 Q1825.45 1550.04 1827.33 1552.34 Q1829.23 1554.63 1829.23 1557.94 Q1829.23 1562.96 1826.15 1565.65 Q1823.09 1568.33 1817.37 1568.33 Q1811.66 1568.33 1808.58 1565.65 Q1805.52 1562.96 1805.52 1557.94 Q1805.52 1554.63 1807.42 1552.34 Q1809.32 1550.04 1812.7 1549.26 M1810.99 1541.87 Q1810.99 1544.56 1812.65 1546.06 Q1814.34 1547.57 1817.37 1547.57 Q1820.38 1547.57 1822.07 1546.06 Q1823.79 1544.56 1823.79 1541.87 Q1823.79 1539.19 1822.07 1537.68 Q1820.38 1536.18 1817.37 1536.18 Q1814.34 1536.18 1812.65 1537.68 Q1810.99 1539.19 1810.99 1541.87 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M2263.76 1563.72 L2271.4 1563.72 L2271.4 1537.36 L2263.09 1539.03 L2263.09 1534.77 L2271.36 1533.1 L2276.03 1533.1 L2276.03 1563.72 L2283.67 1563.72 L2283.67 1567.66 L2263.76 1567.66 L2263.76 1563.72 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M2303.12 1536.18 Q2299.5 1536.18 2297.68 1539.74 Q2295.87 1543.28 2295.87 1550.41 Q2295.87 1557.52 2297.68 1561.09 Q2299.5 1564.63 2303.12 1564.63 Q2306.75 1564.63 2308.56 1561.09 Q2310.38 1557.52 2310.38 1550.41 Q2310.38 1543.28 2308.56 1539.74 Q2306.75 1536.18 2303.12 1536.18 M2303.12 1532.47 Q2308.93 1532.47 2311.98 1537.08 Q2315.06 1541.66 2315.06 1550.41 Q2315.06 1559.14 2311.98 1563.75 Q2308.93 1568.33 2303.12 1568.33 Q2297.31 1568.33 2294.23 1563.75 Q2291.17 1559.14 2291.17 1550.41 Q2291.17 1541.66 2294.23 1537.08 Q2297.31 1532.47 2303.12 1532.47 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M1110.27 1644.78 L1110.27 1663.68 L1104.38 1663.68 L1104.38 1614.48 L1110.27 1614.48 L1110.27 1619.89 Q1112.11 1616.7 1114.92 1615.18 Q1117.75 1613.62 1121.66 1613.62 Q1128.16 1613.62 1132.2 1618.77 Q1136.27 1623.93 1136.27 1632.33 Q1136.27 1640.73 1132.2 1645.89 Q1128.16 1651.05 1121.66 1651.05 Q1117.75 1651.05 1114.92 1649.52 Q1112.11 1647.96 1110.27 1644.78 M1130.19 1632.33 Q1130.19 1625.87 1127.52 1622.21 Q1124.88 1618.52 1120.23 1618.52 Q1115.58 1618.52 1112.91 1622.21 Q1110.27 1625.87 1110.27 1632.33 Q1110.27 1638.79 1112.91 1642.48 Q1115.58 1646.15 1120.23 1646.15 Q1124.88 1646.15 1127.52 1642.48 Q1130.19 1638.79 1130.19 1632.33 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M1159.79 1618.58 Q1155.08 1618.58 1152.35 1622.27 Q1149.61 1625.93 1149.61 1632.33 Q1149.61 1638.73 1152.31 1642.42 Q1155.05 1646.08 1159.79 1646.08 Q1164.47 1646.08 1167.21 1642.39 Q1169.95 1638.7 1169.95 1632.33 Q1169.95 1626 1167.21 1622.31 Q1164.47 1618.58 1159.79 1618.58 M1159.79 1613.62 Q1167.43 1613.62 1171.79 1618.58 Q1176.15 1623.55 1176.15 1632.33 Q1176.15 1641.08 1171.79 1646.08 Q1167.43 1651.05 1159.79 1651.05 Q1152.12 1651.05 1147.76 1646.08 Q1143.43 1641.08 1143.43 1632.33 Q1143.43 1623.55 1147.76 1618.58 Q1152.12 1613.62 1159.79 1613.62 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M1208.59 1615.53 L1208.59 1621.06 Q1206.1 1619.79 1203.43 1619.15 Q1200.76 1618.52 1197.89 1618.52 Q1193.53 1618.52 1191.34 1619.85 Q1189.17 1621.19 1189.17 1623.87 Q1189.17 1625.9 1190.73 1627.08 Q1192.29 1628.23 1197 1629.28 L1199.01 1629.72 Q1205.25 1631.06 1207.86 1633.51 Q1210.5 1635.93 1210.5 1640.29 Q1210.5 1645.25 1206.55 1648.15 Q1202.64 1651.05 1195.76 1651.05 Q1192.9 1651.05 1189.78 1650.47 Q1186.69 1649.93 1183.25 1648.82 L1183.25 1642.77 Q1186.5 1644.46 1189.65 1645.32 Q1192.8 1646.15 1195.89 1646.15 Q1200.03 1646.15 1202.25 1644.74 Q1204.48 1643.31 1204.48 1640.73 Q1204.48 1638.35 1202.86 1637.07 Q1201.27 1635.8 1195.82 1634.62 L1193.79 1634.15 Q1188.34 1633 1185.93 1630.64 Q1183.51 1628.26 1183.51 1624.12 Q1183.51 1619.09 1187.07 1616.35 Q1190.64 1613.62 1197.19 1613.62 Q1200.44 1613.62 1203.3 1614.09 Q1206.17 1614.57 1208.59 1615.53 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M1219.82 1614.48 L1225.68 1614.48 L1225.68 1650.12 L1219.82 1650.12 L1219.82 1614.48 M1219.82 1600.6 L1225.68 1600.6 L1225.68 1608.01 L1219.82 1608.01 L1219.82 1600.6 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M1243.73 1604.35 L1243.73 1614.48 L1255.79 1614.48 L1255.79 1619.03 L1243.73 1619.03 L1243.73 1638.38 Q1243.73 1642.74 1244.9 1643.98 Q1246.11 1645.22 1249.77 1645.22 L1255.79 1645.22 L1255.79 1650.12 L1249.77 1650.12 Q1242.99 1650.12 1240.42 1647.61 Q1237.84 1645.06 1237.84 1638.38 L1237.84 1619.03 L1233.54 1619.03 L1233.54 1614.48 L1237.84 1614.48 L1237.84 1604.35 L1243.73 1604.35 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M1263.49 1614.48 L1269.35 1614.48 L1269.35 1650.12 L1263.49 1650.12 L1263.49 1614.48 M1263.49 1600.6 L1269.35 1600.6 L1269.35 1608.01 L1263.49 1608.01 L1263.49 1600.6 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M1295.42 1618.58 Q1290.7 1618.58 1287.97 1622.27 Q1285.23 1625.93 1285.23 1632.33 Q1285.23 1638.73 1287.94 1642.42 Q1290.67 1646.08 1295.42 1646.08 Q1300.09 1646.08 1302.83 1642.39 Q1305.57 1638.7 1305.57 1632.33 Q1305.57 1626 1302.83 1622.31 Q1300.09 1618.58 1295.42 1618.58 M1295.42 1613.62 Q1303.05 1613.62 1307.41 1618.58 Q1311.78 1623.55 1311.78 1632.33 Q1311.78 1641.08 1307.41 1646.08 Q1303.05 1651.05 1295.42 1651.05 Q1287.74 1651.05 1283.38 1646.08 Q1279.06 1641.08 1279.06 1632.33 Q1279.06 1623.55 1283.38 1618.58 Q1287.74 1613.62 1295.42 1613.62 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M1351.12 1628.61 L1351.12 1650.12 L1345.26 1650.12 L1345.26 1628.8 Q1345.26 1623.74 1343.29 1621.22 Q1341.31 1618.71 1337.37 1618.71 Q1332.62 1618.71 1329.89 1621.73 Q1327.15 1624.76 1327.15 1629.98 L1327.15 1650.12 L1321.26 1650.12 L1321.26 1614.48 L1327.15 1614.48 L1327.15 1620.01 Q1329.25 1616.8 1332.08 1615.21 Q1334.95 1613.62 1338.67 1613.62 Q1344.81 1613.62 1347.96 1617.44 Q1351.12 1621.22 1351.12 1628.61 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><polyline clip-path="url(#clip342)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="102.74,1334.93 2352.76,1334.93 "/>
<polyline clip-path="url(#clip342)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="102.74,977.683 2352.76,977.683 "/>
<polyline clip-path="url(#clip342)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="102.74,620.439 2352.76,620.439 "/>
<polyline clip-path="url(#clip342)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="102.74,263.195 2352.76,263.195 "/>
<polyline clip-path="url(#clip340)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="102.74,1505.26 102.74,47.2441 "/>
<polyline clip-path="url(#clip340)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="102.74,1334.93 121.638,1334.93 "/>
<polyline clip-path="url(#clip340)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="102.74,977.683 121.638,977.683 "/>
<polyline clip-path="url(#clip340)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="102.74,620.439 121.638,620.439 "/>
<polyline clip-path="url(#clip340)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="102.74,263.195 121.638,263.195 "/>
<path clip-path="url(#clip340)" d="M-99.5321 1335.38 L-69.8563 1335.38 L-69.8563 1339.31 L-99.5321 1339.31 L-99.5321 1335.38 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M-49.7638 1320.73 Q-53.3749 1320.73 -55.2036 1324.29 Q-57.0092 1327.83 -57.0092 1334.96 Q-57.0092 1342.07 -55.2036 1345.63 Q-53.3749 1349.17 -49.7638 1349.17 Q-46.1296 1349.17 -44.3241 1345.63 Q-42.4954 1342.07 -42.4954 1334.96 Q-42.4954 1327.83 -44.3241 1324.29 Q-46.1296 1320.73 -49.7638 1320.73 M-49.7638 1317.02 Q-43.9537 1317.02 -40.8982 1321.63 Q-37.8195 1326.21 -37.8195 1334.96 Q-37.8195 1343.69 -40.8982 1348.29 Q-43.9537 1352.88 -49.7638 1352.88 Q-55.574 1352.88 -58.6527 1348.29 Q-61.7082 1343.69 -61.7082 1334.96 Q-61.7082 1326.21 -58.6527 1321.63 Q-55.574 1317.02 -49.7638 1317.02 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M-29.6019 1346.33 L-24.7177 1346.33 L-24.7177 1352.21 L-29.6019 1352.21 L-29.6019 1346.33 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M-4.53261 1320.73 Q-8.1437 1320.73 -9.97239 1324.29 Q-11.7779 1327.83 -11.7779 1334.96 Q-11.7779 1342.07 -9.97239 1345.63 Q-8.1437 1349.17 -4.53261 1349.17 Q-0.898369 1349.17 0.907177 1345.63 Q2.73587 1342.07 2.73587 1334.96 Q2.73587 1327.83 0.907177 1324.29 Q-0.898369 1320.73 -4.53261 1320.73 M-4.53261 1317.02 Q1.27755 1317.02 4.33308 1321.63 Q7.41177 1326.21 7.41177 1334.96 Q7.41177 1343.69 4.33308 1348.29 Q1.27755 1352.88 -4.53261 1352.88 Q-10.3428 1352.88 -13.4215 1348.29 Q-16.477 1343.69 -16.477 1334.96 Q-16.477 1326.21 -13.4215 1321.63 Q-10.3428 1317.02 -4.53261 1317.02 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M16.4395 1348.27 L24.0783 1348.27 L24.0783 1321.91 L15.7682 1323.57 L15.7682 1319.31 L24.0321 1317.65 L28.708 1317.65 L28.708 1348.27 L36.3468 1348.27 L36.3468 1352.21 L16.4395 1352.21 L16.4395 1348.27 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M45.8375 1317.65 L64.1939 1317.65 L64.1939 1321.58 L50.1199 1321.58 L50.1199 1330.05 Q51.1384 1329.71 52.1569 1329.54 Q53.1754 1329.36 54.1939 1329.36 Q59.9809 1329.36 63.3605 1332.53 Q66.7402 1335.7 66.7402 1341.12 Q66.7402 1346.7 63.268 1349.8 Q59.7958 1352.88 53.4763 1352.88 Q51.3004 1352.88 49.0319 1352.51 Q46.7866 1352.14 44.3792 1351.4 L44.3792 1346.7 Q46.4625 1347.83 48.6847 1348.39 Q50.9069 1348.94 53.3837 1348.94 Q57.3884 1348.94 59.7263 1346.84 Q62.0643 1344.73 62.0643 1341.12 Q62.0643 1337.51 59.7263 1335.4 Q57.3884 1333.29 53.3837 1333.29 Q51.5088 1333.29 49.6338 1333.71 Q47.7819 1334.13 45.8375 1335.01 L45.8375 1317.65 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M-100.527 978.134 L-70.8517 978.134 L-70.8517 982.069 L-100.527 982.069 L-100.527 978.134 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M-50.7592 963.482 Q-54.3703 963.482 -56.199 967.046 Q-58.0045 970.588 -58.0045 977.718 Q-58.0045 984.824 -56.199 988.389 Q-54.3703 991.931 -50.7592 991.931 Q-47.125 991.931 -45.3194 988.389 Q-43.4907 984.824 -43.4907 977.718 Q-43.4907 970.588 -45.3194 967.046 Q-47.125 963.482 -50.7592 963.482 M-50.7592 959.778 Q-44.9491 959.778 -41.8935 964.384 Q-38.8148 968.968 -38.8148 977.718 Q-38.8148 986.444 -41.8935 991.051 Q-44.9491 995.634 -50.7592 995.634 Q-56.5694 995.634 -59.6481 991.051 Q-62.7036 986.444 -62.7036 977.718 Q-62.7036 968.968 -59.6481 964.384 Q-56.5694 959.778 -50.7592 959.778 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M-30.5973 989.083 L-25.7131 989.083 L-25.7131 994.963 L-30.5973 994.963 L-30.5973 989.083 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M-5.52797 963.482 Q-9.13907 963.482 -10.9678 967.046 Q-12.7733 970.588 -12.7733 977.718 Q-12.7733 984.824 -10.9678 988.389 Q-9.13907 991.931 -5.52797 991.931 Q-1.89373 991.931 -0.0881881 988.389 Q1.74051 984.824 1.74051 977.718 Q1.74051 970.588 -0.0881881 967.046 Q-1.89373 963.482 -5.52797 963.482 M-5.52797 959.778 Q0.28218 959.778 3.33772 964.384 Q6.41641 968.968 6.41641 977.718 Q6.41641 986.444 3.33772 991.051 Q0.28218 995.634 -5.52797 995.634 Q-11.3381 995.634 -14.4168 991.051 Q-17.4724 986.444 -17.4724 977.718 Q-17.4724 968.968 -14.4168 964.384 Q-11.3381 959.778 -5.52797 959.778 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M15.4441 991.028 L23.083 991.028 L23.083 964.662 L14.7728 966.329 L14.7728 962.07 L23.0367 960.403 L27.7126 960.403 L27.7126 991.028 L35.3514 991.028 L35.3514 994.963 L15.4441 994.963 L15.4441 991.028 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M54.7958 963.482 Q51.1847 963.482 49.356 967.046 Q47.5504 970.588 47.5504 977.718 Q47.5504 984.824 49.356 988.389 Q51.1847 991.931 54.7958 991.931 Q58.43 991.931 60.2356 988.389 Q62.0643 984.824 62.0643 977.718 Q62.0643 970.588 60.2356 967.046 Q58.43 963.482 54.7958 963.482 M54.7958 959.778 Q60.6059 959.778 63.6615 964.384 Q66.7402 968.968 66.7402 977.718 Q66.7402 986.444 63.6615 991.051 Q60.6059 995.634 54.7958 995.634 Q48.9856 995.634 45.9069 991.051 Q42.8514 986.444 42.8514 977.718 Q42.8514 968.968 45.9069 964.384 Q48.9856 959.778 54.7958 959.778 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M-99.5321 620.891 L-69.8563 620.891 L-69.8563 624.826 L-99.5321 624.826 L-99.5321 620.891 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M-49.7638 606.238 Q-53.3749 606.238 -55.2036 609.803 Q-57.0092 613.344 -57.0092 620.474 Q-57.0092 627.58 -55.2036 631.145 Q-53.3749 634.687 -49.7638 634.687 Q-46.1296 634.687 -44.3241 631.145 Q-42.4954 627.58 -42.4954 620.474 Q-42.4954 613.344 -44.3241 609.803 Q-46.1296 606.238 -49.7638 606.238 M-49.7638 602.534 Q-43.9537 602.534 -40.8982 607.141 Q-37.8195 611.724 -37.8195 620.474 Q-37.8195 629.201 -40.8982 633.807 Q-43.9537 638.39 -49.7638 638.39 Q-55.574 638.39 -58.6527 633.807 Q-61.7082 629.201 -61.7082 620.474 Q-61.7082 611.724 -58.6527 607.141 Q-55.574 602.534 -49.7638 602.534 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M-29.6019 631.84 L-24.7177 631.84 L-24.7177 637.719 L-29.6019 637.719 L-29.6019 631.84 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M-4.53261 606.238 Q-8.1437 606.238 -9.97239 609.803 Q-11.7779 613.344 -11.7779 620.474 Q-11.7779 627.58 -9.97239 631.145 Q-8.1437 634.687 -4.53261 634.687 Q-0.898369 634.687 0.907177 631.145 Q2.73587 627.58 2.73587 620.474 Q2.73587 613.344 0.907177 609.803 Q-0.898369 606.238 -4.53261 606.238 M-4.53261 602.534 Q1.27755 602.534 4.33308 607.141 Q7.41177 611.724 7.41177 620.474 Q7.41177 629.201 4.33308 633.807 Q1.27755 638.39 -4.53261 638.39 Q-10.3428 638.39 -13.4215 633.807 Q-16.477 629.201 -16.477 620.474 Q-16.477 611.724 -13.4215 607.141 Q-10.3428 602.534 -4.53261 602.534 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M25.6293 606.238 Q22.0182 606.238 20.1895 609.803 Q18.3839 613.344 18.3839 620.474 Q18.3839 627.58 20.1895 631.145 Q22.0182 634.687 25.6293 634.687 Q29.2635 634.687 31.0691 631.145 Q32.8977 627.58 32.8977 620.474 Q32.8977 613.344 31.0691 609.803 Q29.2635 606.238 25.6293 606.238 M25.6293 602.534 Q31.4394 602.534 34.495 607.141 Q37.5736 611.724 37.5736 620.474 Q37.5736 629.201 34.495 633.807 Q31.4394 638.39 25.6293 638.39 Q19.8191 638.39 16.7404 633.807 Q13.6849 629.201 13.6849 620.474 Q13.6849 611.724 16.7404 607.141 Q19.8191 602.534 25.6293 602.534 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M45.8375 603.159 L64.1939 603.159 L64.1939 607.094 L50.1199 607.094 L50.1199 615.567 Q51.1384 615.219 52.1569 615.057 Q53.1754 614.872 54.1939 614.872 Q59.9809 614.872 63.3605 618.043 Q66.7402 621.215 66.7402 626.631 Q66.7402 632.21 63.268 635.312 Q59.7958 638.39 53.4763 638.39 Q51.3004 638.39 49.0319 638.02 Q46.7866 637.65 44.3792 636.909 L44.3792 632.21 Q46.4625 633.344 48.6847 633.9 Q50.9069 634.455 53.3837 634.455 Q57.3884 634.455 59.7263 632.349 Q62.0643 630.242 62.0643 626.631 Q62.0643 623.02 59.7263 620.914 Q57.3884 618.807 53.3837 618.807 Q51.5088 618.807 49.6338 619.224 Q47.7819 619.641 45.8375 620.52 L45.8375 603.159 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M-50.7592 248.994 Q-54.3703 248.994 -56.199 252.559 Q-58.0045 256.101 -58.0045 263.23 Q-58.0045 270.337 -56.199 273.901 Q-54.3703 277.443 -50.7592 277.443 Q-47.125 277.443 -45.3194 273.901 Q-43.4907 270.337 -43.4907 263.23 Q-43.4907 256.101 -45.3194 252.559 Q-47.125 248.994 -50.7592 248.994 M-50.7592 245.29 Q-44.9491 245.29 -41.8935 249.897 Q-38.8148 254.48 -38.8148 263.23 Q-38.8148 271.957 -41.8935 276.563 Q-44.9491 281.147 -50.7592 281.147 Q-56.5694 281.147 -59.6481 276.563 Q-62.7036 271.957 -62.7036 263.23 Q-62.7036 254.48 -59.6481 249.897 Q-56.5694 245.29 -50.7592 245.29 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M-30.5973 274.596 L-25.7131 274.596 L-25.7131 280.475 L-30.5973 280.475 L-30.5973 274.596 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M-5.52797 248.994 Q-9.13907 248.994 -10.9678 252.559 Q-12.7733 256.101 -12.7733 263.23 Q-12.7733 270.337 -10.9678 273.901 Q-9.13907 277.443 -5.52797 277.443 Q-1.89373 277.443 -0.0881881 273.901 Q1.74051 270.337 1.74051 263.23 Q1.74051 256.101 -0.0881881 252.559 Q-1.89373 248.994 -5.52797 248.994 M-5.52797 245.29 Q0.28218 245.29 3.33772 249.897 Q6.41641 254.48 6.41641 263.23 Q6.41641 271.957 3.33772 276.563 Q0.28218 281.147 -5.52797 281.147 Q-11.3381 281.147 -14.4168 276.563 Q-17.4724 271.957 -17.4724 263.23 Q-17.4724 254.48 -14.4168 249.897 Q-11.3381 245.29 -5.52797 245.29 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M24.6339 248.994 Q21.0228 248.994 19.1941 252.559 Q17.3886 256.101 17.3886 263.23 Q17.3886 270.337 19.1941 273.901 Q21.0228 277.443 24.6339 277.443 Q28.2681 277.443 30.0737 273.901 Q31.9024 270.337 31.9024 263.23 Q31.9024 256.101 30.0737 252.559 Q28.2681 248.994 24.6339 248.994 M24.6339 245.29 Q30.4441 245.29 33.4996 249.897 Q36.5783 254.48 36.5783 263.23 Q36.5783 271.957 33.4996 276.563 Q30.4441 281.147 24.6339 281.147 Q18.8237 281.147 15.7451 276.563 Q12.6895 271.957 12.6895 263.23 Q12.6895 254.48 15.7451 249.897 Q18.8237 245.29 24.6339 245.29 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M54.7958 248.994 Q51.1847 248.994 49.356 252.559 Q47.5504 256.101 47.5504 263.23 Q47.5504 270.337 49.356 273.901 Q51.1847 277.443 54.7958 277.443 Q58.43 277.443 60.2356 273.901 Q62.0643 270.337 62.0643 263.23 Q62.0643 256.101 60.2356 252.559 Q58.43 248.994 54.7958 248.994 M54.7958 245.29 Q60.6059 245.29 63.6615 249.897 Q66.7402 254.48 66.7402 263.23 Q66.7402 271.957 63.6615 276.563 Q60.6059 281.147 54.7958 281.147 Q48.9856 281.147 45.9069 276.563 Q42.8514 271.957 42.8514 263.23 Q42.8514 254.48 45.9069 249.897 Q48.9856 245.29 54.7958 245.29 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M-170.072 983.869 L-167.207 983.869 L-167.207 1010.8 Q-161.16 1010.41 -157.977 1007.17 Q-154.826 1003.89 -154.826 998.065 Q-154.826 994.691 -155.653 991.54 Q-156.481 988.357 -158.136 985.238 L-152.598 985.238 Q-151.261 988.389 -150.561 991.699 Q-149.86 995.009 -149.86 998.415 Q-149.86 1006.95 -154.826 1011.94 Q-159.791 1016.91 -168.257 1016.91 Q-177.01 1016.91 -182.135 1012.2 Q-187.291 1007.45 -187.291 999.434 Q-187.291 992.24 -182.644 988.071 Q-178.029 983.869 -170.072 983.869 M-171.79 989.726 Q-176.596 989.79 -179.461 992.431 Q-182.326 995.041 -182.326 999.37 Q-182.326 1004.27 -179.556 1007.23 Q-176.787 1010.16 -171.758 1010.61 L-171.79 989.726 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M-172.3 944.625 L-150.783 944.625 L-150.783 950.481 L-172.109 950.481 Q-177.169 950.481 -179.684 952.455 Q-182.198 954.428 -182.198 958.375 Q-182.198 963.117 -179.175 965.855 Q-176.151 968.592 -170.931 968.592 L-150.783 968.592 L-150.783 974.48 L-186.431 974.48 L-186.431 968.592 L-180.893 968.592 Q-184.108 966.491 -185.699 963.658 Q-187.291 960.794 -187.291 957.07 Q-187.291 950.927 -183.471 947.776 Q-179.684 944.625 -172.3 944.625 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M-170.072 902.452 L-167.207 902.452 L-167.207 929.379 Q-161.16 928.997 -157.977 925.751 Q-154.826 922.472 -154.826 916.648 Q-154.826 913.274 -155.653 910.123 Q-156.481 906.94 -158.136 903.821 L-152.598 903.821 Q-151.261 906.972 -150.561 910.282 Q-149.86 913.592 -149.86 916.998 Q-149.86 925.528 -154.826 930.525 Q-159.791 935.49 -168.257 935.49 Q-177.01 935.49 -182.135 930.779 Q-187.291 926.037 -187.291 918.016 Q-187.291 910.823 -182.644 906.653 Q-178.029 902.452 -170.072 902.452 M-171.79 908.309 Q-176.596 908.372 -179.461 911.014 Q-182.326 913.624 -182.326 917.953 Q-182.326 922.854 -179.556 925.814 Q-176.787 928.742 -171.758 929.188 L-171.79 908.309 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M-180.957 872.183 Q-181.53 873.17 -181.784 874.348 Q-182.071 875.493 -182.071 876.894 Q-182.071 881.859 -178.824 884.533 Q-175.61 887.174 -169.562 887.174 L-150.783 887.174 L-150.783 893.063 L-186.431 893.063 L-186.431 887.174 L-180.893 887.174 Q-184.14 885.328 -185.699 882.368 Q-187.291 879.408 -187.291 875.175 Q-187.291 874.57 -187.195 873.838 Q-187.132 873.106 -186.973 872.215 L-180.957 872.183 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M-169.021 843.728 Q-175.387 843.728 -178.888 846.37 Q-182.389 848.98 -182.389 853.723 Q-182.389 858.433 -178.888 861.075 Q-175.387 863.685 -169.021 863.685 Q-162.687 863.685 -159.186 861.075 Q-155.685 858.433 -155.685 853.723 Q-155.685 848.98 -159.186 846.37 Q-162.687 843.728 -169.021 843.728 M-155.208 837.872 Q-146.105 837.872 -141.681 841.914 Q-137.225 845.956 -137.225 854.296 Q-137.225 857.383 -137.702 860.12 Q-138.148 862.857 -139.102 865.436 L-144.8 865.436 Q-143.399 862.857 -142.731 860.343 Q-142.062 857.828 -142.062 855.219 Q-142.062 849.458 -145.086 846.593 Q-148.078 843.728 -154.157 843.728 L-157.054 843.728 Q-153.903 845.543 -152.343 848.375 Q-150.783 851.208 -150.783 855.155 Q-150.783 861.712 -155.781 865.722 Q-160.778 869.732 -169.021 869.732 Q-177.297 869.732 -182.294 865.722 Q-187.291 861.712 -187.291 855.155 Q-187.291 851.208 -185.731 848.375 Q-184.172 845.543 -181.021 843.728 L-186.431 843.728 L-186.431 837.872 L-155.208 837.872 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M-147.473 810.977 Q-141.108 813.46 -139.166 815.815 Q-137.225 818.17 -137.225 822.117 L-137.225 826.796 L-142.126 826.796 L-142.126 823.358 Q-142.126 820.939 -143.272 819.602 Q-144.418 818.266 -148.683 816.642 L-151.356 815.592 L-186.431 830.01 L-186.431 823.804 L-158.55 812.664 L-186.431 801.524 L-186.431 795.317 L-147.473 810.977 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M-181.021 743.055 L-200.309 743.055 L-200.309 737.198 L-150.783 737.198 L-150.783 743.055 L-156.131 743.055 Q-152.948 744.901 -151.388 747.734 Q-149.86 750.535 -149.86 754.481 Q-149.86 760.942 -155.017 765.016 Q-160.173 769.059 -168.576 769.059 Q-176.978 769.059 -182.135 765.016 Q-187.291 760.942 -187.291 754.481 Q-187.291 750.535 -185.731 747.734 Q-184.203 744.901 -181.021 743.055 M-168.576 763.011 Q-162.114 763.011 -158.422 760.37 Q-154.762 757.696 -154.762 753.049 Q-154.762 748.402 -158.422 745.728 Q-162.114 743.055 -168.576 743.055 Q-175.037 743.055 -178.697 745.728 Q-182.389 748.402 -182.389 753.049 Q-182.389 757.696 -178.697 760.37 Q-175.037 763.011 -168.576 763.011 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M-170.072 694.644 L-167.207 694.644 L-167.207 721.571 Q-161.16 721.189 -157.977 717.942 Q-154.826 714.664 -154.826 708.839 Q-154.826 705.465 -155.653 702.314 Q-156.481 699.131 -158.136 696.012 L-152.598 696.012 Q-151.261 699.163 -150.561 702.473 Q-149.86 705.784 -149.86 709.189 Q-149.86 717.719 -154.826 722.716 Q-159.791 727.682 -168.257 727.682 Q-177.01 727.682 -182.135 722.971 Q-187.291 718.229 -187.291 710.208 Q-187.291 703.015 -182.644 698.845 Q-178.029 694.644 -170.072 694.644 M-171.79 700.5 Q-176.596 700.564 -179.461 703.205 Q-182.326 705.815 -182.326 710.144 Q-182.326 715.046 -179.556 718.006 Q-176.787 720.934 -171.758 721.38 L-171.79 700.5 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M-172.3 655.399 L-150.783 655.399 L-150.783 661.255 L-172.109 661.255 Q-177.169 661.255 -179.684 663.229 Q-182.198 665.202 -182.198 669.149 Q-182.198 673.891 -179.175 676.629 Q-176.151 679.366 -170.931 679.366 L-150.783 679.366 L-150.783 685.254 L-186.431 685.254 L-186.431 679.366 L-180.893 679.366 Q-184.108 677.265 -185.699 674.432 Q-187.291 671.568 -187.291 667.844 Q-187.291 661.701 -183.471 658.55 Q-179.684 655.399 -172.3 655.399 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M-185.381 620.992 L-179.843 620.992 Q-181.116 623.475 -181.753 626.149 Q-182.389 628.822 -182.389 631.687 Q-182.389 636.047 -181.052 638.243 Q-179.716 640.408 -177.042 640.408 Q-175.005 640.408 -173.827 638.848 Q-172.681 637.289 -171.631 632.578 L-171.186 630.573 Q-169.849 624.334 -167.398 621.724 Q-164.979 619.083 -160.618 619.083 Q-155.653 619.083 -152.757 623.029 Q-149.86 626.944 -149.86 633.819 Q-149.86 636.684 -150.433 639.803 Q-150.974 642.89 -152.088 646.328 L-158.136 646.328 Q-156.449 643.081 -155.59 639.93 Q-154.762 636.779 -154.762 633.692 Q-154.762 629.554 -156.162 627.326 Q-157.595 625.098 -160.173 625.098 Q-162.56 625.098 -163.833 626.722 Q-165.106 628.313 -166.284 633.756 L-166.761 635.793 Q-167.907 641.235 -170.263 643.654 Q-172.65 646.073 -176.787 646.073 Q-181.816 646.073 -184.554 642.508 Q-187.291 638.944 -187.291 632.387 Q-187.291 629.14 -186.813 626.276 Q-186.336 623.411 -185.381 620.992 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M-186.431 609.757 L-186.431 603.9 L-150.783 603.9 L-150.783 609.757 L-186.431 609.757 M-200.309 609.757 L-200.309 603.9 L-192.893 603.9 L-192.893 609.757 L-200.309 609.757 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M-196.553 585.854 L-186.431 585.854 L-186.431 573.791 L-181.88 573.791 L-181.88 585.854 L-162.528 585.854 Q-158.168 585.854 -156.926 584.676 Q-155.685 583.467 -155.685 579.806 L-155.685 573.791 L-150.783 573.791 L-150.783 579.806 Q-150.783 586.586 -153.298 589.164 Q-155.844 591.742 -162.528 591.742 L-181.88 591.742 L-181.88 596.039 L-186.431 596.039 L-186.431 591.742 L-196.553 591.742 L-196.553 585.854 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M-147.473 551.256 Q-141.108 553.739 -139.166 556.094 Q-137.225 558.449 -137.225 562.396 L-137.225 567.075 L-142.126 567.075 L-142.126 563.637 Q-142.126 561.218 -143.272 559.882 Q-144.418 558.545 -148.683 556.922 L-151.356 555.871 L-186.431 570.29 L-186.431 564.083 L-158.55 552.943 L-186.431 541.803 L-186.431 535.596 L-147.473 551.256 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><polyline clip-path="url(#clip342)" style="stroke:#009af9; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="166.42,407.535 402.271,890.003 638.121,1030.99 873.972,492.43 1109.82,106.255 1345.67,106.255 1581.52,492.43 1817.37,1030.99 2053.23,890.003 2289.08,407.535 "/>
<polyline clip-path="url(#clip342)" style="stroke:#e26f46; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="166.42,489.669 402.271,1214.36 638.121,1464 873.972,705.722 1109.82,88.5087 1345.67,88.5087 1581.52,705.722 1817.37,1464 2053.23,1214.36 2289.08,489.669 "/>
<path clip-path="url(#clip340)" d="M1930.14 251.365 L2277.76 251.365 L2277.76 95.8446 L1930.14 95.8446  Z" fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<polyline clip-path="url(#clip340)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="1930.14,251.365 2277.76,251.365 2277.76,95.8446 1930.14,95.8446 1930.14,251.365 "/>
<polyline clip-path="url(#clip340)" style="stroke:#009af9; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="1955.14,147.685 2105.14,147.685 "/>
<path clip-path="url(#clip340)" d="M2152.39 131.539 L2152.39 136.099 Q2149.72 134.826 2147.36 134.201 Q2145 133.576 2142.8 133.576 Q2138.98 133.576 2136.9 135.057 Q2134.84 136.539 2134.84 139.27 Q2134.84 141.562 2136.2 142.743 Q2137.59 143.9 2141.44 144.618 L2144.26 145.196 Q2149.49 146.192 2151.97 148.715 Q2154.47 151.215 2154.47 155.428 Q2154.47 160.451 2151.09 163.043 Q2147.73 165.636 2141.23 165.636 Q2138.77 165.636 2136 165.08 Q2133.24 164.525 2130.28 163.437 L2130.28 158.622 Q2133.13 160.219 2135.86 161.029 Q2138.59 161.84 2141.23 161.84 Q2145.23 161.84 2147.41 160.266 Q2149.58 158.692 2149.58 155.775 Q2149.58 153.229 2148.01 151.793 Q2146.46 150.358 2142.89 149.641 L2140.05 149.085 Q2134.82 148.043 2132.48 145.821 Q2130.14 143.599 2130.14 139.641 Q2130.14 135.057 2133.36 132.418 Q2136.6 129.78 2142.27 129.78 Q2144.7 129.78 2147.22 130.219 Q2149.75 130.659 2152.39 131.539 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M2177.2 143.437 L2206.88 143.437 L2206.88 147.326 L2177.2 147.326 L2177.2 143.437 M2177.2 152.881 L2206.88 152.881 L2206.88 156.817 L2177.2 156.817 L2177.2 152.881 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M2232.85 161.029 L2240.49 161.029 L2240.49 134.664 L2232.18 136.331 L2232.18 132.071 L2240.44 130.405 L2245.12 130.405 L2245.12 161.029 L2252.76 161.029 L2252.76 164.965 L2232.85 164.965 L2232.85 161.029 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><polyline clip-path="url(#clip340)" style="stroke:#e26f46; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="1955.14,199.525 2105.14,199.525 "/>
<path clip-path="url(#clip340)" d="M2152.39 183.379 L2152.39 187.939 Q2149.72 186.666 2147.36 186.041 Q2145 185.416 2142.8 185.416 Q2138.98 185.416 2136.9 186.897 Q2134.84 188.379 2134.84 191.11 Q2134.84 193.402 2136.2 194.583 Q2137.59 195.74 2141.44 196.458 L2144.26 197.036 Q2149.49 198.032 2151.97 200.555 Q2154.47 203.055 2154.47 207.268 Q2154.47 212.291 2151.09 214.883 Q2147.73 217.476 2141.23 217.476 Q2138.77 217.476 2136 216.92 Q2133.24 216.365 2130.28 215.277 L2130.28 210.462 Q2133.13 212.059 2135.86 212.869 Q2138.59 213.68 2141.23 213.68 Q2145.23 213.68 2147.41 212.106 Q2149.58 210.532 2149.58 207.615 Q2149.58 205.069 2148.01 203.633 Q2146.46 202.198 2142.89 201.481 L2140.05 200.925 Q2134.82 199.883 2132.48 197.661 Q2130.14 195.439 2130.14 191.481 Q2130.14 186.897 2133.36 184.258 Q2136.6 181.62 2142.27 181.62 Q2144.7 181.62 2147.22 182.059 Q2149.75 182.499 2152.39 183.379 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M2177.2 195.277 L2206.88 195.277 L2206.88 199.166 L2177.2 199.166 L2177.2 195.277 M2177.2 204.721 L2206.88 204.721 L2206.88 208.657 L2177.2 208.657 L2177.2 204.721 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip340)" d="M2236.07 212.869 L2252.38 212.869 L2252.38 216.805 L2230.44 216.805 L2230.44 212.869 Q2233.1 210.115 2237.69 205.485 Q2242.29 200.832 2243.47 199.49 Q2245.72 196.967 2246.6 195.231 Q2247.5 193.471 2247.5 191.782 Q2247.5 189.027 2245.56 187.291 Q2243.63 185.555 2240.53 185.555 Q2238.33 185.555 2235.88 186.319 Q2233.45 187.083 2230.67 188.633 L2230.67 183.911 Q2233.5 182.777 2235.95 182.198 Q2238.4 181.62 2240.44 181.62 Q2245.81 181.62 2249.01 184.305 Q2252.2 186.99 2252.2 191.481 Q2252.2 193.61 2251.39 195.532 Q2250.6 197.43 2248.5 200.022 Q2247.92 200.694 2244.82 203.911 Q2241.71 207.106 2236.07 212.869 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /></svg>

```

Finally, we can obtain a value for the Haldane gap by extrapolating our results for different system sizes.

````julia
Ls = 5:5:50
ΔEs = map(Ls) do L
    ψ₀ = FiniteMPS(rand, ComplexF64, L, physical_space, virtual_space)
    ψ, envs, delta = find_groundstate(ψ₀, H, DMRG(; verbose=false))
    En_1, st_1 = excitations(H, QuasiparticleAnsatz(), ψ, envs; sector=SU₂(1))
    En_2, st_2 = excitations(H, QuasiparticleAnsatz(), ψ, envs; sector=SU₂(2))
    return real(En_2[1] - En_1[1])
end

f = fit(Ls .^ (-2), ΔEs, 1)
ΔE_extrapolated = f.coeffs[1]
````

````
-0.2155532516993712
````

````julia
p_size_extrapolation = plot(; xaxis="L^(-2)", yaxis="ΔE",
    xlims=(0, 0.015))
plot!(p_size_extrapolation, Ls .^ (-2), ΔEs;
    seriestype=:scatter, label="numerical")
plot!(p_size_extrapolation, x -> f(x); label="fit")
````

```@raw html
<?xml version="1.0" encoding="utf-8"?>
<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" width="600" height="400" viewBox="0 0 2400 1600">
<defs>
  <clipPath id="clip370">
    <rect x="0" y="0" width="2400" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip370)" d="M0 1600 L2400 1600 L2400 0 L0 0  Z" fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip371">
    <rect x="480" y="0" width="1681" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip370)" d="M102.74 1505.26 L2352.76 1505.26 L2352.76 47.2441 L102.74 47.2441  Z" fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip372">
    <rect x="102" y="47" width="2251" height="1459"/>
  </clipPath>
</defs>
<polyline clip-path="url(#clip372)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="102.74,1505.26 102.74,47.2441 "/>
<polyline clip-path="url(#clip372)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="852.745,1505.26 852.745,47.2441 "/>
<polyline clip-path="url(#clip372)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="1602.75,1505.26 1602.75,47.2441 "/>
<polyline clip-path="url(#clip372)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="2352.76,1505.26 2352.76,47.2441 "/>
<polyline clip-path="url(#clip370)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="102.74,1505.26 2352.76,1505.26 "/>
<polyline clip-path="url(#clip370)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="102.74,1505.26 102.74,1486.36 "/>
<polyline clip-path="url(#clip370)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="852.745,1505.26 852.745,1486.36 "/>
<polyline clip-path="url(#clip370)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="1602.75,1505.26 1602.75,1486.36 "/>
<polyline clip-path="url(#clip370)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="2352.76,1505.26 2352.76,1486.36 "/>
<path clip-path="url(#clip370)" d="M49.9627 1536.18 Q46.3516 1536.18 44.5229 1539.74 Q42.7173 1543.28 42.7173 1550.41 Q42.7173 1557.52 44.5229 1561.09 Q46.3516 1564.63 49.9627 1564.63 Q53.5969 1564.63 55.4024 1561.09 Q57.2311 1557.52 57.2311 1550.41 Q57.2311 1543.28 55.4024 1539.74 Q53.5969 1536.18 49.9627 1536.18 M49.9627 1532.47 Q55.7728 1532.47 58.8284 1537.08 Q61.907 1541.66 61.907 1550.41 Q61.907 1559.14 58.8284 1563.75 Q55.7728 1568.33 49.9627 1568.33 Q44.1525 1568.33 41.0738 1563.75 Q38.0183 1559.14 38.0183 1550.41 Q38.0183 1541.66 41.0738 1537.08 Q44.1525 1532.47 49.9627 1532.47 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M70.1246 1561.78 L75.0088 1561.78 L75.0088 1567.66 L70.1246 1567.66 L70.1246 1561.78 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M95.1939 1536.18 Q91.5828 1536.18 89.7541 1539.74 Q87.9486 1543.28 87.9486 1550.41 Q87.9486 1557.52 89.7541 1561.09 Q91.5828 1564.63 95.1939 1564.63 Q98.8281 1564.63 100.634 1561.09 Q102.462 1557.52 102.462 1550.41 Q102.462 1543.28 100.634 1539.74 Q98.8281 1536.18 95.1939 1536.18 M95.1939 1532.47 Q101.004 1532.47 104.06 1537.08 Q107.138 1541.66 107.138 1550.41 Q107.138 1559.14 104.06 1563.75 Q101.004 1568.33 95.1939 1568.33 Q89.3837 1568.33 86.3051 1563.75 Q83.2495 1559.14 83.2495 1550.41 Q83.2495 1541.66 86.3051 1537.08 Q89.3837 1532.47 95.1939 1532.47 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M125.356 1536.18 Q121.745 1536.18 119.916 1539.74 Q118.11 1543.28 118.11 1550.41 Q118.11 1557.52 119.916 1561.09 Q121.745 1564.63 125.356 1564.63 Q128.99 1564.63 130.796 1561.09 Q132.624 1557.52 132.624 1550.41 Q132.624 1543.28 130.796 1539.74 Q128.99 1536.18 125.356 1536.18 M125.356 1532.47 Q131.166 1532.47 134.221 1537.08 Q137.3 1541.66 137.3 1550.41 Q137.3 1559.14 134.221 1563.75 Q131.166 1568.33 125.356 1568.33 Q119.546 1568.33 116.467 1563.75 Q113.411 1559.14 113.411 1550.41 Q113.411 1541.66 116.467 1537.08 Q119.546 1532.47 125.356 1532.47 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M155.518 1536.18 Q151.907 1536.18 150.078 1539.74 Q148.272 1543.28 148.272 1550.41 Q148.272 1557.52 150.078 1561.09 Q151.907 1564.63 155.518 1564.63 Q159.152 1564.63 160.957 1561.09 Q162.786 1557.52 162.786 1550.41 Q162.786 1543.28 160.957 1539.74 Q159.152 1536.18 155.518 1536.18 M155.518 1532.47 Q161.328 1532.47 164.383 1537.08 Q167.462 1541.66 167.462 1550.41 Q167.462 1559.14 164.383 1563.75 Q161.328 1568.33 155.518 1568.33 Q149.707 1568.33 146.629 1563.75 Q143.573 1559.14 143.573 1550.41 Q143.573 1541.66 146.629 1537.08 Q149.707 1532.47 155.518 1532.47 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M800.466 1536.18 Q796.855 1536.18 795.026 1539.74 Q793.22 1543.28 793.22 1550.41 Q793.22 1557.52 795.026 1561.09 Q796.855 1564.63 800.466 1564.63 Q804.1 1564.63 805.905 1561.09 Q807.734 1557.52 807.734 1550.41 Q807.734 1543.28 805.905 1539.74 Q804.1 1536.18 800.466 1536.18 M800.466 1532.47 Q806.276 1532.47 809.331 1537.08 Q812.41 1541.66 812.41 1550.41 Q812.41 1559.14 809.331 1563.75 Q806.276 1568.33 800.466 1568.33 Q794.655 1568.33 791.577 1563.75 Q788.521 1559.14 788.521 1550.41 Q788.521 1541.66 791.577 1537.08 Q794.655 1532.47 800.466 1532.47 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M820.628 1561.78 L825.512 1561.78 L825.512 1567.66 L820.628 1567.66 L820.628 1561.78 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M845.697 1536.18 Q842.086 1536.18 840.257 1539.74 Q838.452 1543.28 838.452 1550.41 Q838.452 1557.52 840.257 1561.09 Q842.086 1564.63 845.697 1564.63 Q849.331 1564.63 851.137 1561.09 Q852.965 1557.52 852.965 1550.41 Q852.965 1543.28 851.137 1539.74 Q849.331 1536.18 845.697 1536.18 M845.697 1532.47 Q851.507 1532.47 854.563 1537.08 Q857.641 1541.66 857.641 1550.41 Q857.641 1559.14 854.563 1563.75 Q851.507 1568.33 845.697 1568.33 Q839.887 1568.33 836.808 1563.75 Q833.752 1559.14 833.752 1550.41 Q833.752 1541.66 836.808 1537.08 Q839.887 1532.47 845.697 1532.47 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M875.859 1536.18 Q872.248 1536.18 870.419 1539.74 Q868.613 1543.28 868.613 1550.41 Q868.613 1557.52 870.419 1561.09 Q872.248 1564.63 875.859 1564.63 Q879.493 1564.63 881.298 1561.09 Q883.127 1557.52 883.127 1550.41 Q883.127 1543.28 881.298 1539.74 Q879.493 1536.18 875.859 1536.18 M875.859 1532.47 Q881.669 1532.47 884.724 1537.08 Q887.803 1541.66 887.803 1550.41 Q887.803 1559.14 884.724 1563.75 Q881.669 1568.33 875.859 1568.33 Q870.049 1568.33 866.97 1563.75 Q863.914 1559.14 863.914 1550.41 Q863.914 1541.66 866.97 1537.08 Q870.049 1532.47 875.859 1532.47 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M896.067 1533.1 L914.423 1533.1 L914.423 1537.04 L900.349 1537.04 L900.349 1545.51 Q901.368 1545.16 902.386 1545 Q903.405 1544.81 904.423 1544.81 Q910.21 1544.81 913.59 1547.98 Q916.97 1551.16 916.97 1556.57 Q916.97 1562.15 913.497 1565.25 Q910.025 1568.33 903.706 1568.33 Q901.53 1568.33 899.261 1567.96 Q897.016 1567.59 894.609 1566.85 L894.609 1562.15 Q896.692 1563.28 898.914 1563.84 Q901.136 1564.4 903.613 1564.4 Q907.618 1564.4 909.956 1562.29 Q912.294 1560.18 912.294 1556.57 Q912.294 1552.96 909.956 1550.85 Q907.618 1548.75 903.613 1548.75 Q901.738 1548.75 899.863 1549.16 Q898.011 1549.58 896.067 1550.46 L896.067 1533.1 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M1549.97 1536.18 Q1546.36 1536.18 1544.53 1539.74 Q1542.73 1543.28 1542.73 1550.41 Q1542.73 1557.52 1544.53 1561.09 Q1546.36 1564.63 1549.97 1564.63 Q1553.61 1564.63 1555.41 1561.09 Q1557.24 1557.52 1557.24 1550.41 Q1557.24 1543.28 1555.41 1539.74 Q1553.61 1536.18 1549.97 1536.18 M1549.97 1532.47 Q1555.78 1532.47 1558.84 1537.08 Q1561.92 1541.66 1561.92 1550.41 Q1561.92 1559.14 1558.84 1563.75 Q1555.78 1568.33 1549.97 1568.33 Q1544.16 1568.33 1541.08 1563.75 Q1538.03 1559.14 1538.03 1550.41 Q1538.03 1541.66 1541.08 1537.08 Q1544.16 1532.47 1549.97 1532.47 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M1570.14 1561.78 L1575.02 1561.78 L1575.02 1567.66 L1570.14 1567.66 L1570.14 1561.78 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M1595.2 1536.18 Q1591.59 1536.18 1589.76 1539.74 Q1587.96 1543.28 1587.96 1550.41 Q1587.96 1557.52 1589.76 1561.09 Q1591.59 1564.63 1595.2 1564.63 Q1598.84 1564.63 1600.64 1561.09 Q1602.47 1557.52 1602.47 1550.41 Q1602.47 1543.28 1600.64 1539.74 Q1598.84 1536.18 1595.2 1536.18 M1595.2 1532.47 Q1601.01 1532.47 1604.07 1537.08 Q1607.15 1541.66 1607.15 1550.41 Q1607.15 1559.14 1604.07 1563.75 Q1601.01 1568.33 1595.2 1568.33 Q1589.39 1568.33 1586.32 1563.75 Q1583.26 1559.14 1583.26 1550.41 Q1583.26 1541.66 1586.32 1537.08 Q1589.39 1532.47 1595.2 1532.47 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M1616.18 1563.72 L1623.82 1563.72 L1623.82 1537.36 L1615.51 1539.03 L1615.51 1534.77 L1623.77 1533.1 L1628.44 1533.1 L1628.44 1563.72 L1636.08 1563.72 L1636.08 1567.66 L1616.18 1567.66 L1616.18 1563.72 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M1655.53 1536.18 Q1651.92 1536.18 1650.09 1539.74 Q1648.28 1543.28 1648.28 1550.41 Q1648.28 1557.52 1650.09 1561.09 Q1651.92 1564.63 1655.53 1564.63 Q1659.16 1564.63 1660.97 1561.09 Q1662.8 1557.52 1662.8 1550.41 Q1662.8 1543.28 1660.97 1539.74 Q1659.16 1536.18 1655.53 1536.18 M1655.53 1532.47 Q1661.34 1532.47 1664.39 1537.08 Q1667.47 1541.66 1667.47 1550.41 Q1667.47 1559.14 1664.39 1563.75 Q1661.34 1568.33 1655.53 1568.33 Q1649.72 1568.33 1646.64 1563.75 Q1643.58 1559.14 1643.58 1550.41 Q1643.58 1541.66 1646.64 1537.08 Q1649.72 1532.47 1655.53 1532.47 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M2300.48 1536.18 Q2296.87 1536.18 2295.04 1539.74 Q2293.23 1543.28 2293.23 1550.41 Q2293.23 1557.52 2295.04 1561.09 Q2296.87 1564.63 2300.48 1564.63 Q2304.11 1564.63 2305.92 1561.09 Q2307.74 1557.52 2307.74 1550.41 Q2307.74 1543.28 2305.92 1539.74 Q2304.11 1536.18 2300.48 1536.18 M2300.48 1532.47 Q2306.29 1532.47 2309.34 1537.08 Q2312.42 1541.66 2312.42 1550.41 Q2312.42 1559.14 2309.34 1563.75 Q2306.29 1568.33 2300.48 1568.33 Q2294.67 1568.33 2291.59 1563.75 Q2288.53 1559.14 2288.53 1550.41 Q2288.53 1541.66 2291.59 1537.08 Q2294.67 1532.47 2300.48 1532.47 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M2320.64 1561.78 L2325.52 1561.78 L2325.52 1567.66 L2320.64 1567.66 L2320.64 1561.78 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M2345.71 1536.18 Q2342.1 1536.18 2340.27 1539.74 Q2338.46 1543.28 2338.46 1550.41 Q2338.46 1557.52 2340.27 1561.09 Q2342.1 1564.63 2345.71 1564.63 Q2349.34 1564.63 2351.15 1561.09 Q2352.98 1557.52 2352.98 1550.41 Q2352.98 1543.28 2351.15 1539.74 Q2349.34 1536.18 2345.71 1536.18 M2345.71 1532.47 Q2351.52 1532.47 2354.57 1537.08 Q2357.65 1541.66 2357.65 1550.41 Q2357.65 1559.14 2354.57 1563.75 Q2351.52 1568.33 2345.71 1568.33 Q2339.9 1568.33 2336.82 1563.75 Q2333.76 1559.14 2333.76 1550.41 Q2333.76 1541.66 2336.82 1537.08 Q2339.9 1532.47 2345.71 1532.47 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M2366.68 1563.72 L2374.32 1563.72 L2374.32 1537.36 L2366.01 1539.03 L2366.01 1534.77 L2374.27 1533.1 L2378.95 1533.1 L2378.95 1563.72 L2386.59 1563.72 L2386.59 1567.66 L2366.68 1567.66 L2366.68 1563.72 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M2396.08 1533.1 L2414.43 1533.1 L2414.43 1537.04 L2400.36 1537.04 L2400.36 1545.51 Q2401.38 1545.16 2402.4 1545 Q2403.42 1544.81 2404.43 1544.81 Q2410.22 1544.81 2413.6 1547.98 Q2416.98 1551.16 2416.98 1556.57 Q2416.98 1562.15 2413.51 1565.25 Q2410.04 1568.33 2403.72 1568.33 Q2401.54 1568.33 2399.27 1567.96 Q2397.03 1567.59 2394.62 1566.85 L2394.62 1562.15 Q2396.7 1563.28 2398.92 1563.84 Q2401.15 1564.4 2403.62 1564.4 Q2407.63 1564.4 2409.97 1562.29 Q2412.3 1560.18 2412.3 1556.57 Q2412.3 1552.96 2409.97 1550.85 Q2407.63 1548.75 2403.62 1548.75 Q2401.75 1548.75 2399.87 1549.16 Q2398.02 1549.58 2396.08 1550.46 L2396.08 1533.1 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M1130.35 1602.6 L1136.78 1602.6 L1136.78 1644.71 L1159.92 1644.71 L1159.92 1650.12 L1130.35 1650.12 L1130.35 1602.6 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M1190.7 1602.6 L1207.98 1620.33 L1201.58 1620.33 L1187.58 1607.76 L1173.58 1620.33 L1167.18 1620.33 L1184.46 1602.6 L1190.7 1602.6 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M1235.1 1600.66 Q1230.84 1607.98 1228.77 1615.14 Q1226.7 1622.31 1226.7 1629.66 Q1226.7 1637.01 1228.77 1644.24 Q1230.87 1651.43 1235.1 1658.72 L1230.01 1658.72 Q1225.23 1651.24 1222.85 1644.01 Q1220.49 1636.79 1220.49 1629.66 Q1220.49 1622.56 1222.85 1615.37 Q1225.2 1608.17 1230.01 1600.66 L1235.1 1600.66 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M1243.5 1629.66 L1260.66 1629.66 L1260.66 1634.88 L1243.5 1634.88 L1243.5 1629.66 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M1276.35 1644.71 L1298.79 1644.71 L1298.79 1650.12 L1268.62 1650.12 L1268.62 1644.71 Q1272.28 1640.93 1278.58 1634.56 Q1284.91 1628.16 1286.54 1626.32 Q1289.62 1622.85 1290.83 1620.46 Q1292.07 1618.04 1292.07 1615.72 Q1292.07 1611.93 1289.4 1609.54 Q1286.76 1607.16 1282.49 1607.16 Q1279.47 1607.16 1276.1 1608.21 Q1272.75 1609.26 1268.93 1611.39 L1268.93 1604.9 Q1272.82 1603.34 1276.19 1602.54 Q1279.56 1601.74 1282.37 1601.74 Q1289.75 1601.74 1294.14 1605.44 Q1298.53 1609.13 1298.53 1615.3 Q1298.53 1618.23 1297.42 1620.87 Q1296.34 1623.48 1293.44 1627.05 Q1292.65 1627.97 1288.38 1632.4 Q1284.12 1636.79 1276.35 1644.71 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M1310.53 1600.66 L1315.63 1600.66 Q1320.4 1608.17 1322.76 1615.37 Q1325.14 1622.56 1325.14 1629.66 Q1325.14 1636.79 1322.76 1644.01 Q1320.4 1651.24 1315.63 1658.72 L1310.53 1658.72 Q1314.77 1651.43 1316.84 1644.24 Q1318.94 1637.01 1318.94 1629.66 Q1318.94 1622.31 1316.84 1615.14 Q1314.77 1607.98 1310.53 1600.66 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><polyline clip-path="url(#clip372)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="102.74,1334.42 2352.76,1334.42 "/>
<polyline clip-path="url(#clip372)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="102.74,1080.2 2352.76,1080.2 "/>
<polyline clip-path="url(#clip372)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="102.74,825.986 2352.76,825.986 "/>
<polyline clip-path="url(#clip372)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="102.74,571.769 2352.76,571.769 "/>
<polyline clip-path="url(#clip372)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="102.74,317.551 2352.76,317.551 "/>
<polyline clip-path="url(#clip372)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="102.74,63.3328 2352.76,63.3328 "/>
<polyline clip-path="url(#clip370)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="102.74,1505.26 102.74,47.2441 "/>
<polyline clip-path="url(#clip370)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="102.74,1334.42 121.638,1334.42 "/>
<polyline clip-path="url(#clip370)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="102.74,1080.2 121.638,1080.2 "/>
<polyline clip-path="url(#clip370)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="102.74,825.986 121.638,825.986 "/>
<polyline clip-path="url(#clip370)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="102.74,571.769 121.638,571.769 "/>
<polyline clip-path="url(#clip370)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="102.74,317.551 121.638,317.551 "/>
<polyline clip-path="url(#clip370)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="102.74,63.3328 121.638,63.3328 "/>
<path clip-path="url(#clip370)" d="M-69.3702 1334.87 L-39.6945 1334.87 L-39.6945 1338.81 L-69.3702 1338.81 L-69.3702 1334.87 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M-19.602 1320.22 Q-23.2131 1320.22 -25.0418 1323.79 Q-26.8473 1327.33 -26.8473 1334.46 Q-26.8473 1341.56 -25.0418 1345.13 Q-23.2131 1348.67 -19.602 1348.67 Q-15.9677 1348.67 -14.1622 1345.13 Q-12.3335 1341.56 -12.3335 1334.46 Q-12.3335 1327.33 -14.1622 1323.79 Q-15.9677 1320.22 -19.602 1320.22 M-19.602 1316.52 Q-13.7918 1316.52 -10.7363 1321.12 Q-7.65759 1325.71 -7.65759 1334.46 Q-7.65759 1343.18 -10.7363 1347.79 Q-13.7918 1352.37 -19.602 1352.37 Q-25.4121 1352.37 -28.4908 1347.79 Q-31.5464 1343.18 -31.5464 1334.46 Q-31.5464 1325.71 -28.4908 1321.12 Q-25.4121 1316.52 -19.602 1316.52 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M0.559957 1345.82 L5.44419 1345.82 L5.44419 1351.7 L0.559957 1351.7 L0.559957 1345.82 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M28.4765 1321.22 L16.671 1339.67 L28.4765 1339.67 L28.4765 1321.22 M27.2496 1317.14 L33.1292 1317.14 L33.1292 1339.67 L38.0598 1339.67 L38.0598 1343.55 L33.1292 1343.55 L33.1292 1351.7 L28.4765 1351.7 L28.4765 1343.55 L12.8747 1343.55 L12.8747 1339.04 L27.2496 1317.14 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M45.8375 1317.14 L64.1939 1317.14 L64.1939 1321.08 L50.1199 1321.08 L50.1199 1329.55 Q51.1384 1329.2 52.1569 1329.04 Q53.1754 1328.86 54.1939 1328.86 Q59.9809 1328.86 63.3605 1332.03 Q66.7402 1335.2 66.7402 1340.61 Q66.7402 1346.19 63.268 1349.29 Q59.7958 1352.37 53.4763 1352.37 Q51.3004 1352.37 49.0319 1352 Q46.7866 1351.63 44.3792 1350.89 L44.3792 1346.19 Q46.4625 1347.33 48.6847 1347.88 Q50.9069 1348.44 53.3837 1348.44 Q57.3884 1348.44 59.7263 1346.33 Q62.0643 1344.23 62.0643 1340.61 Q62.0643 1337 59.7263 1334.9 Q57.3884 1332.79 53.3837 1332.79 Q51.5088 1332.79 49.6338 1333.21 Q47.7819 1333.62 45.8375 1334.5 L45.8375 1317.14 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M-70.3656 1080.66 L-40.6898 1080.66 L-40.6898 1084.59 L-70.3656 1084.59 L-70.3656 1080.66 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M-20.5973 1066 Q-24.2084 1066 -26.0371 1069.57 Q-27.8427 1073.11 -27.8427 1080.24 Q-27.8427 1087.35 -26.0371 1090.91 Q-24.2084 1094.45 -20.5973 1094.45 Q-16.9631 1094.45 -15.1576 1090.91 Q-13.3289 1087.35 -13.3289 1080.24 Q-13.3289 1073.11 -15.1576 1069.57 Q-16.9631 1066 -20.5973 1066 M-20.5973 1062.3 Q-14.7872 1062.3 -11.7316 1066.91 Q-8.65296 1071.49 -8.65296 1080.24 Q-8.65296 1088.97 -11.7316 1093.57 Q-14.7872 1098.16 -20.5973 1098.16 Q-26.4075 1098.16 -29.4862 1093.57 Q-32.5417 1088.97 -32.5417 1080.24 Q-32.5417 1071.49 -29.4862 1066.91 Q-26.4075 1062.3 -20.5973 1062.3 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M-0.435408 1091.6 L4.44882 1091.6 L4.44882 1097.48 L-0.435408 1097.48 L-0.435408 1091.6 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M27.4811 1067 L15.6756 1085.45 L27.4811 1085.45 L27.4811 1067 M26.2543 1062.92 L32.1339 1062.92 L32.1339 1085.45 L37.0644 1085.45 L37.0644 1089.34 L32.1339 1089.34 L32.1339 1097.48 L27.4811 1097.48 L27.4811 1089.34 L11.8793 1089.34 L11.8793 1084.82 L26.2543 1062.92 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M54.7958 1066 Q51.1847 1066 49.356 1069.57 Q47.5504 1073.11 47.5504 1080.24 Q47.5504 1087.35 49.356 1090.91 Q51.1847 1094.45 54.7958 1094.45 Q58.43 1094.45 60.2356 1090.91 Q62.0643 1087.35 62.0643 1080.24 Q62.0643 1073.11 60.2356 1069.57 Q58.43 1066 54.7958 1066 M54.7958 1062.3 Q60.6059 1062.3 63.6615 1066.91 Q66.7402 1071.49 66.7402 1080.24 Q66.7402 1088.97 63.6615 1093.57 Q60.6059 1098.16 54.7958 1098.16 Q48.9856 1098.16 45.9069 1093.57 Q42.8514 1088.97 42.8514 1080.24 Q42.8514 1071.49 45.9069 1066.91 Q48.9856 1062.3 54.7958 1062.3 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M-69.3702 826.438 L-39.6945 826.438 L-39.6945 830.373 L-69.3702 830.373 L-69.3702 826.438 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M-19.602 811.785 Q-23.2131 811.785 -25.0418 815.35 Q-26.8473 818.892 -26.8473 826.021 Q-26.8473 833.128 -25.0418 836.692 Q-23.2131 840.234 -19.602 840.234 Q-15.9677 840.234 -14.1622 836.692 Q-12.3335 833.128 -12.3335 826.021 Q-12.3335 818.892 -14.1622 815.35 Q-15.9677 811.785 -19.602 811.785 M-19.602 808.081 Q-13.7918 808.081 -10.7363 812.688 Q-7.65759 817.271 -7.65759 826.021 Q-7.65759 834.748 -10.7363 839.354 Q-13.7918 843.938 -19.602 843.938 Q-25.4121 843.938 -28.4908 839.354 Q-31.5464 834.748 -31.5464 826.021 Q-31.5464 817.271 -28.4908 812.688 Q-25.4121 808.081 -19.602 808.081 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M0.559957 837.387 L5.44419 837.387 L5.44419 843.266 L0.559957 843.266 L0.559957 837.387 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M29.7959 824.632 Q33.1524 825.35 35.0274 827.618 Q36.9255 829.887 36.9255 833.22 Q36.9255 838.336 33.407 841.137 Q29.8885 843.938 23.4071 843.938 Q21.2311 843.938 18.9163 843.498 Q16.6247 843.081 14.171 842.225 L14.171 837.711 Q16.1154 838.845 18.4302 839.424 Q20.745 840.003 23.2682 840.003 Q27.6663 840.003 29.9579 838.266 Q32.2727 836.53 32.2727 833.22 Q32.2727 830.165 30.12 828.452 Q27.9904 826.716 24.1709 826.716 L20.1432 826.716 L20.1432 822.873 L24.3561 822.873 Q27.8052 822.873 29.6339 821.507 Q31.4626 820.118 31.4626 817.526 Q31.4626 814.864 29.5644 813.452 Q27.6894 812.017 24.1709 812.017 Q22.2497 812.017 20.0506 812.433 Q17.8515 812.85 15.2127 813.73 L15.2127 809.563 Q17.8747 808.822 20.1895 808.452 Q22.5274 808.081 24.5876 808.081 Q29.9117 808.081 33.0135 810.512 Q36.1153 812.919 36.1153 817.04 Q36.1153 819.91 34.4718 821.901 Q32.8283 823.868 29.7959 824.632 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M45.8375 808.706 L64.1939 808.706 L64.1939 812.642 L50.1199 812.642 L50.1199 821.114 Q51.1384 820.767 52.1569 820.605 Q53.1754 820.419 54.1939 820.419 Q59.9809 820.419 63.3605 823.591 Q66.7402 826.762 66.7402 832.179 Q66.7402 837.757 63.268 840.859 Q59.7958 843.938 53.4763 843.938 Q51.3004 843.938 49.0319 843.567 Q46.7866 843.197 44.3792 842.456 L44.3792 837.757 Q46.4625 838.891 48.6847 839.447 Q50.9069 840.003 53.3837 840.003 Q57.3884 840.003 59.7263 837.896 Q62.0643 835.79 62.0643 832.179 Q62.0643 828.567 59.7263 826.461 Q57.3884 824.355 53.3837 824.355 Q51.5088 824.355 49.6338 824.771 Q47.7819 825.188 45.8375 826.067 L45.8375 808.706 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M-70.3656 572.22 L-40.6898 572.22 L-40.6898 576.155 L-70.3656 576.155 L-70.3656 572.22 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M-20.5973 557.567 Q-24.2084 557.567 -26.0371 561.132 Q-27.8427 564.674 -27.8427 571.803 Q-27.8427 578.91 -26.0371 582.475 Q-24.2084 586.016 -20.5973 586.016 Q-16.9631 586.016 -15.1576 582.475 Q-13.3289 578.91 -13.3289 571.803 Q-13.3289 564.674 -15.1576 561.132 Q-16.9631 557.567 -20.5973 557.567 M-20.5973 553.864 Q-14.7872 553.864 -11.7316 558.47 Q-8.65296 563.053 -8.65296 571.803 Q-8.65296 580.53 -11.7316 585.137 Q-14.7872 589.72 -20.5973 589.72 Q-26.4075 589.72 -29.4862 585.137 Q-32.5417 580.53 -32.5417 571.803 Q-32.5417 563.053 -29.4862 558.47 Q-26.4075 553.864 -20.5973 553.864 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M-0.435408 583.169 L4.44882 583.169 L4.44882 589.049 L-0.435408 589.049 L-0.435408 583.169 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M28.8005 570.414 Q32.157 571.132 34.032 573.401 Q35.9301 575.669 35.9301 579.002 Q35.9301 584.118 32.4116 586.919 Q28.8931 589.72 22.4117 589.72 Q20.2358 589.72 17.921 589.28 Q15.6293 588.863 13.1756 588.007 L13.1756 583.493 Q15.1201 584.627 17.4349 585.206 Q19.7497 585.785 22.2728 585.785 Q26.6709 585.785 28.9626 584.049 Q31.2774 582.313 31.2774 579.002 Q31.2774 575.947 29.1246 574.234 Q26.995 572.498 23.1756 572.498 L19.1478 572.498 L19.1478 568.655 L23.3608 568.655 Q26.8098 568.655 28.6385 567.289 Q30.4672 565.901 30.4672 563.308 Q30.4672 560.646 28.5691 559.234 Q26.6941 557.799 23.1756 557.799 Q21.2543 557.799 19.0552 558.215 Q16.8562 558.632 14.2173 559.512 L14.2173 555.345 Q16.8793 554.604 19.1941 554.234 Q21.5321 553.864 23.5922 553.864 Q28.9163 553.864 32.0181 556.294 Q35.12 558.702 35.12 562.822 Q35.12 565.692 33.4764 567.683 Q31.8329 569.651 28.8005 570.414 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M54.7958 557.567 Q51.1847 557.567 49.356 561.132 Q47.5504 564.674 47.5504 571.803 Q47.5504 578.91 49.356 582.475 Q51.1847 586.016 54.7958 586.016 Q58.43 586.016 60.2356 582.475 Q62.0643 578.91 62.0643 571.803 Q62.0643 564.674 60.2356 561.132 Q58.43 557.567 54.7958 557.567 M54.7958 553.864 Q60.6059 553.864 63.6615 558.47 Q66.7402 563.053 66.7402 571.803 Q66.7402 580.53 63.6615 585.137 Q60.6059 589.72 54.7958 589.72 Q48.9856 589.72 45.9069 585.137 Q42.8514 580.53 42.8514 571.803 Q42.8514 563.053 45.9069 558.47 Q48.9856 553.864 54.7958 553.864 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M-69.3702 318.002 L-39.6945 318.002 L-39.6945 321.937 L-69.3702 321.937 L-69.3702 318.002 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M-19.602 303.349 Q-23.2131 303.349 -25.0418 306.914 Q-26.8473 310.456 -26.8473 317.585 Q-26.8473 324.692 -25.0418 328.257 Q-23.2131 331.798 -19.602 331.798 Q-15.9677 331.798 -14.1622 328.257 Q-12.3335 324.692 -12.3335 317.585 Q-12.3335 310.456 -14.1622 306.914 Q-15.9677 303.349 -19.602 303.349 M-19.602 299.646 Q-13.7918 299.646 -10.7363 304.252 Q-7.65759 308.835 -7.65759 317.585 Q-7.65759 326.312 -10.7363 330.919 Q-13.7918 335.502 -19.602 335.502 Q-25.4121 335.502 -28.4908 330.919 Q-31.5464 326.312 -31.5464 317.585 Q-31.5464 308.835 -28.4908 304.252 Q-25.4121 299.646 -19.602 299.646 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M0.559957 328.951 L5.44419 328.951 L5.44419 334.831 L0.559957 334.831 L0.559957 328.951 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M19.6571 330.896 L35.9764 330.896 L35.9764 334.831 L14.0321 334.831 L14.0321 330.896 Q16.6941 328.141 21.2774 323.511 Q25.8839 318.859 27.0644 317.516 Q29.3098 314.993 30.1894 313.257 Q31.0922 311.497 31.0922 309.808 Q31.0922 307.053 29.1478 305.317 Q27.2265 303.581 24.1246 303.581 Q21.9256 303.581 19.4719 304.345 Q17.0413 305.109 14.2636 306.66 L14.2636 301.937 Q17.0876 300.803 19.5413 300.224 Q21.995 299.646 24.0321 299.646 Q29.4024 299.646 32.5968 302.331 Q35.7912 305.016 35.7912 309.507 Q35.7912 311.636 34.9811 313.558 Q34.194 315.456 32.0876 318.048 Q31.5089 318.72 28.407 321.937 Q25.3052 325.132 19.6571 330.896 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M45.8375 300.271 L64.1939 300.271 L64.1939 304.206 L50.1199 304.206 L50.1199 312.678 Q51.1384 312.331 52.1569 312.169 Q53.1754 311.984 54.1939 311.984 Q59.9809 311.984 63.3605 315.155 Q66.7402 318.326 66.7402 323.743 Q66.7402 329.321 63.268 332.423 Q59.7958 335.502 53.4763 335.502 Q51.3004 335.502 49.0319 335.132 Q46.7866 334.761 44.3792 334.021 L44.3792 329.321 Q46.4625 330.456 48.6847 331.011 Q50.9069 331.567 53.3837 331.567 Q57.3884 331.567 59.7263 329.46 Q62.0643 327.354 62.0643 323.743 Q62.0643 320.132 59.7263 318.025 Q57.3884 315.919 53.3837 315.919 Q51.5088 315.919 49.6338 316.335 Q47.7819 316.752 45.8375 317.632 L45.8375 300.271 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M-70.3656 63.7842 L-40.6898 63.7842 L-40.6898 67.7194 L-70.3656 67.7194 L-70.3656 63.7842 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M-20.5973 49.1315 Q-24.2084 49.1315 -26.0371 52.6963 Q-27.8427 56.238 -27.8427 63.3675 Q-27.8427 70.474 -26.0371 74.0388 Q-24.2084 77.5804 -20.5973 77.5804 Q-16.9631 77.5804 -15.1576 74.0388 Q-13.3289 70.474 -13.3289 63.3675 Q-13.3289 56.238 -15.1576 52.6963 Q-16.9631 49.1315 -20.5973 49.1315 M-20.5973 45.4278 Q-14.7872 45.4278 -11.7316 50.0343 Q-8.65296 54.6176 -8.65296 63.3675 Q-8.65296 72.0943 -11.7316 76.7008 Q-14.7872 81.2841 -20.5973 81.2841 Q-26.4075 81.2841 -29.4862 76.7008 Q-32.5417 72.0943 -32.5417 63.3675 Q-32.5417 54.6176 -29.4862 50.0343 Q-26.4075 45.4278 -20.5973 45.4278 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M-0.435408 74.7332 L4.44882 74.7332 L4.44882 80.6128 L-0.435408 80.6128 L-0.435408 74.7332 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M18.6617 76.6777 L34.9811 76.6777 L34.9811 80.6128 L13.0367 80.6128 L13.0367 76.6777 Q15.6988 73.923 20.2821 69.2934 Q24.8885 64.6407 26.0691 63.2981 Q28.3144 60.775 29.1941 59.0389 Q30.0968 57.2796 30.0968 55.5898 Q30.0968 52.8352 28.1524 51.0991 Q26.2311 49.363 23.1293 49.363 Q20.9302 49.363 18.4765 50.1269 Q16.046 50.8908 13.2682 52.4417 L13.2682 47.7195 Q16.0923 46.5852 18.546 46.0065 Q20.9997 45.4278 23.0367 45.4278 Q28.407 45.4278 31.6015 48.113 Q34.7959 50.7982 34.7959 55.2889 Q34.7959 57.4185 33.9857 59.3398 Q33.1987 61.2379 31.0922 63.8305 Q30.5135 64.5018 27.4117 67.7194 Q24.3098 70.9138 18.6617 76.6777 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M54.7958 49.1315 Q51.1847 49.1315 49.356 52.6963 Q47.5504 56.238 47.5504 63.3675 Q47.5504 70.474 49.356 74.0388 Q51.1847 77.5804 54.7958 77.5804 Q58.43 77.5804 60.2356 74.0388 Q62.0643 70.474 62.0643 63.3675 Q62.0643 56.238 60.2356 52.6963 Q58.43 49.1315 54.7958 49.1315 M54.7958 45.4278 Q60.6059 45.4278 63.6615 50.0343 Q66.7402 54.6176 66.7402 63.3675 Q66.7402 72.0943 63.6615 76.7008 Q60.6059 81.2841 54.7958 81.2841 Q48.9856 81.2841 45.9069 76.7008 Q42.8514 72.0943 42.8514 63.3675 Q42.8514 54.6176 45.9069 50.0343 Q48.9856 45.4278 54.7958 45.4278 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M-161.808 795.031 L-125.969 808.081 L-125.969 781.949 L-161.808 795.031 M-120.622 816.802 L-168.142 798.659 L-168.142 791.371 L-120.622 773.26 L-120.622 816.802 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M-168.142 766.321 L-168.142 736.275 L-162.731 736.275 L-162.731 759.892 L-148.663 759.892 L-148.663 737.262 L-143.252 737.262 L-143.252 759.892 L-126.032 759.892 L-126.032 735.702 L-120.622 735.702 L-120.622 766.321 L-168.142 766.321 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><circle clip-path="url(#clip372)" cx="1602.75" cy="519.553" r="14.4" fill="#009af9" fill-rule="evenodd" fill-opacity="1" stroke="#000000" stroke-opacity="1" stroke-width="3.2"/>
<circle clip-path="url(#clip372)" cx="769.411" cy="96.6463" r="14.4" fill="#009af9" fill-rule="evenodd" fill-opacity="1" stroke="#000000" stroke-opacity="1" stroke-width="3.2"/>
<circle clip-path="url(#clip372)" cx="477.743" cy="223.106" r="14.4" fill="#009af9" fill-rule="evenodd" fill-opacity="1" stroke="#000000" stroke-opacity="1" stroke-width="3.2"/>
<circle clip-path="url(#clip372)" cx="342.742" cy="294.061" r="14.4" fill="#009af9" fill-rule="evenodd" fill-opacity="1" stroke="#000000" stroke-opacity="1" stroke-width="3.2"/>
<circle clip-path="url(#clip372)" cx="269.408" cy="88.5087" r="14.4" fill="#009af9" fill-rule="evenodd" fill-opacity="1" stroke="#000000" stroke-opacity="1" stroke-width="3.2"/>
<circle clip-path="url(#clip372)" cx="225.19" cy="228.546" r="14.4" fill="#009af9" fill-rule="evenodd" fill-opacity="1" stroke="#000000" stroke-opacity="1" stroke-width="3.2"/>
<circle clip-path="url(#clip372)" cx="196.491" cy="181.856" r="14.4" fill="#009af9" fill-rule="evenodd" fill-opacity="1" stroke="#000000" stroke-opacity="1" stroke-width="3.2"/>
<circle clip-path="url(#clip372)" cx="176.815" cy="169.051" r="14.4" fill="#009af9" fill-rule="evenodd" fill-opacity="1" stroke="#000000" stroke-opacity="1" stroke-width="3.2"/>
<circle clip-path="url(#clip372)" cx="162.741" cy="197.844" r="14.4" fill="#009af9" fill-rule="evenodd" fill-opacity="1" stroke="#000000" stroke-opacity="1" stroke-width="3.2"/>
<polyline clip-path="url(#clip372)" style="stroke:#e26f46; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="102.74,142.411 117.389,145.623 252.839,175.325 334.238,193.175 403.06,208.267 477.821,224.661 546.743,239.774 620.635,255.978 706.246,274.751 782.328,291.434 849.044,306.064 921.783,322.015 994.954,338.06 1075.85,355.8 1152.83,372.679 1232.86,390.23 1300.24,405.004 1382.49,423.041 1457.84,439.564 1527.43,454.824 1597.93,470.283 1676.93,487.607 1748.52,503.306 1830.49,521.28 1906.38,537.922 1981.2,554.329 2053.85,570.26 2128.32,586.59 2209.91,604.482 2328.44,630.473 2352.76,635.805 "/>
<path clip-path="url(#clip370)" d="M1825.12 251.365 L2277.76 251.365 L2277.76 95.8446 L1825.12 95.8446  Z" fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<polyline clip-path="url(#clip370)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="1825.12,251.365 2277.76,251.365 2277.76,95.8446 1825.12,95.8446 1825.12,251.365 "/>
<circle clip-path="url(#clip370)" cx="1925.12" cy="147.685" r="20.48" fill="#009af9" fill-rule="evenodd" fill-opacity="1" stroke="#000000" stroke-opacity="1" stroke-width="4.55111"/>
<path clip-path="url(#clip370)" d="M2046.83 149.317 L2046.83 164.965 L2042.57 164.965 L2042.57 149.455 Q2042.57 145.775 2041.14 143.946 Q2039.7 142.118 2036.83 142.118 Q2033.38 142.118 2031.39 144.317 Q2029.4 146.516 2029.4 150.312 L2029.4 164.965 L2025.12 164.965 L2025.12 139.039 L2029.4 139.039 L2029.4 143.067 Q2030.93 140.729 2032.99 139.571 Q2035.07 138.414 2037.78 138.414 Q2042.25 138.414 2044.54 141.192 Q2046.83 143.946 2046.83 149.317 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M2054.89 154.733 L2054.89 139.039 L2059.15 139.039 L2059.15 154.571 Q2059.15 158.252 2060.58 160.104 Q2062.02 161.932 2064.89 161.932 Q2068.33 161.932 2070.33 159.733 Q2072.34 157.534 2072.34 153.738 L2072.34 139.039 L2076.6 139.039 L2076.6 164.965 L2072.34 164.965 L2072.34 160.983 Q2070.79 163.344 2068.73 164.502 Q2066.69 165.636 2063.98 165.636 Q2059.52 165.636 2057.2 162.858 Q2054.89 160.08 2054.89 154.733 M2065.6 138.414 L2065.6 138.414 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M2105.56 144.016 Q2107.15 141.145 2109.38 139.78 Q2111.6 138.414 2114.61 138.414 Q2118.66 138.414 2120.86 141.261 Q2123.06 144.085 2123.06 149.317 L2123.06 164.965 L2118.77 164.965 L2118.77 149.455 Q2118.77 145.729 2117.46 143.923 Q2116.14 142.118 2113.43 142.118 Q2110.12 142.118 2108.2 144.317 Q2106.27 146.516 2106.27 150.312 L2106.27 164.965 L2101.99 164.965 L2101.99 149.455 Q2101.99 145.705 2100.67 143.923 Q2099.35 142.118 2096.6 142.118 Q2093.33 142.118 2091.41 144.34 Q2089.49 146.539 2089.49 150.312 L2089.49 164.965 L2085.21 164.965 L2085.21 139.039 L2089.49 139.039 L2089.49 143.067 Q2090.95 140.682 2092.99 139.548 Q2095.02 138.414 2097.83 138.414 Q2100.65 138.414 2102.62 139.849 Q2104.61 141.284 2105.56 144.016 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M2153.73 150.937 L2153.73 153.02 L2134.14 153.02 Q2134.42 157.418 2136.78 159.733 Q2139.17 162.025 2143.4 162.025 Q2145.86 162.025 2148.15 161.423 Q2150.46 160.821 2152.73 159.617 L2152.73 163.645 Q2150.44 164.617 2148.03 165.127 Q2145.63 165.636 2143.15 165.636 Q2136.95 165.636 2133.31 162.025 Q2129.7 158.414 2129.7 152.256 Q2129.7 145.891 2133.13 142.164 Q2136.58 138.414 2142.41 138.414 Q2147.64 138.414 2150.67 141.793 Q2153.73 145.15 2153.73 150.937 M2149.47 149.687 Q2149.42 146.192 2147.5 144.108 Q2145.6 142.025 2142.45 142.025 Q2138.89 142.025 2136.74 144.039 Q2134.61 146.053 2134.28 149.71 L2149.47 149.687 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M2175.74 143.02 Q2175.02 142.604 2174.17 142.418 Q2173.33 142.21 2172.32 142.21 Q2168.7 142.21 2166.76 144.571 Q2164.84 146.909 2164.84 151.307 L2164.84 164.965 L2160.56 164.965 L2160.56 139.039 L2164.84 139.039 L2164.84 143.067 Q2166.18 140.705 2168.33 139.571 Q2170.49 138.414 2173.57 138.414 Q2174.01 138.414 2174.54 138.483 Q2175.07 138.53 2175.72 138.645 L2175.74 143.02 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M2180.21 139.039 L2184.47 139.039 L2184.47 164.965 L2180.21 164.965 L2180.21 139.039 M2180.21 128.946 L2184.47 128.946 L2184.47 134.34 L2180.21 134.34 L2180.21 128.946 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M2212.04 140.034 L2212.04 144.016 Q2210.23 143.02 2208.4 142.534 Q2206.6 142.025 2204.75 142.025 Q2200.6 142.025 2198.31 144.664 Q2196.02 147.28 2196.02 152.025 Q2196.02 156.77 2198.31 159.409 Q2200.6 162.025 2204.75 162.025 Q2206.6 162.025 2208.4 161.539 Q2210.23 161.029 2212.04 160.034 L2212.04 163.969 Q2210.26 164.803 2208.33 165.219 Q2206.44 165.636 2204.28 165.636 Q2198.43 165.636 2194.98 161.955 Q2191.53 158.275 2191.53 152.025 Q2191.53 145.682 2195 142.048 Q2198.5 138.414 2204.56 138.414 Q2206.53 138.414 2208.4 138.831 Q2210.28 139.224 2212.04 140.034 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M2231.23 151.932 Q2226.07 151.932 2224.07 153.113 Q2222.08 154.293 2222.08 157.141 Q2222.08 159.409 2223.57 160.752 Q2225.07 162.071 2227.64 162.071 Q2231.18 162.071 2233.31 159.571 Q2235.46 157.048 2235.46 152.881 L2235.46 151.932 L2231.23 151.932 M2239.72 150.173 L2239.72 164.965 L2235.46 164.965 L2235.46 161.029 Q2234.01 163.391 2231.83 164.525 Q2229.65 165.636 2226.51 165.636 Q2222.52 165.636 2220.16 163.414 Q2217.82 161.168 2217.82 157.418 Q2217.82 153.043 2220.74 150.821 Q2223.68 148.599 2229.49 148.599 L2235.46 148.599 L2235.46 148.182 Q2235.46 145.243 2233.52 143.645 Q2231.6 142.025 2228.1 142.025 Q2225.88 142.025 2223.77 142.557 Q2221.67 143.09 2219.72 144.155 L2219.72 140.219 Q2222.06 139.317 2224.26 138.877 Q2226.46 138.414 2228.54 138.414 Q2234.17 138.414 2236.95 141.33 Q2239.72 144.247 2239.72 150.173 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M2248.5 128.946 L2252.76 128.946 L2252.76 164.965 L2248.5 164.965 L2248.5 128.946 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><polyline clip-path="url(#clip370)" style="stroke:#e26f46; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="1850.12,199.525 2000.12,199.525 "/>
<path clip-path="url(#clip370)" d="M2041.62 180.786 L2041.62 184.328 L2037.55 184.328 Q2035.26 184.328 2034.35 185.254 Q2033.47 186.18 2033.47 188.587 L2033.47 190.879 L2040.49 190.879 L2040.49 194.189 L2033.47 194.189 L2033.47 216.805 L2029.19 216.805 L2029.19 194.189 L2025.12 194.189 L2025.12 190.879 L2029.19 190.879 L2029.19 189.073 Q2029.19 184.745 2031.21 182.777 Q2033.22 180.786 2037.59 180.786 L2041.62 180.786 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M2045.19 190.879 L2049.45 190.879 L2049.45 216.805 L2045.19 216.805 L2045.19 190.879 M2045.19 180.786 L2049.45 180.786 L2049.45 186.18 L2045.19 186.18 L2045.19 180.786 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip370)" d="M2062.57 183.518 L2062.57 190.879 L2071.34 190.879 L2071.34 194.189 L2062.57 194.189 L2062.57 208.263 Q2062.57 211.434 2063.43 212.337 Q2064.31 213.24 2066.97 213.24 L2071.34 213.24 L2071.34 216.805 L2066.97 216.805 Q2062.04 216.805 2060.16 214.976 Q2058.29 213.124 2058.29 208.263 L2058.29 194.189 L2055.16 194.189 L2055.16 190.879 L2058.29 190.879 L2058.29 183.518 L2062.57 183.518 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /></svg>

```

## Thermodynamic limit

A much nicer way of obtaining the Haldane gap is by working directly in the thermodynamic limit.
As was already hinted at by the edge modes, this model is in a non-trivial SPT phase.
Thus, care must be taken when selecting the symmetry sectors.
The groundstate has half-integer edge modes, thus the virtual spaces must also all carry half-integer charges.

In contrast with the finite size case, we now should specify a momentum label to the excitations.
This way, it is possible to scan the dispersion relation over the entire momentum space.

````julia
virtual_space_inf = Rep[SU₂](1 // 2 => 20, 3 // 2 => 20, 5 // 2 => 10, 7 // 2 => 10,
    9 // 2 => 5)
ψ₀_inf = InfiniteMPS([physical_space], [virtual_space_inf])
ψ_inf, envs_inf, delta_inf = find_groundstate(ψ₀_inf, H)

kspace = range(0, π, 16)
Es, _ = excitations(H, QuasiparticleAnsatz(), kspace, ψ_inf, envs_inf; sector=SU₂(1))

ΔE, idx = findmin(real.(Es))
println("minimum @k = $(kspace[idx]):\t ΔE = $(ΔE)")
````

````
[ Info: vumps @iteration 1 galerkin = 0.6648889732866761
[ Info: vumps @iteration 2 galerkin = 0.06359011437557995
[ Info: vumps @iteration 3 galerkin = 0.0164265217276693
[ Info: vumps @iteration 4 galerkin = 0.002373916187780843
[ Info: vumps @iteration 5 galerkin = 0.0006289175273869992
[ Info: vumps @iteration 6 galerkin = 0.00022983228453123588
[ Info: vumps @iteration 7 galerkin = 0.00010467216817002606
[ Info: vumps @iteration 8 galerkin = 5.438736021221384e-5
[ Info: CG: initializing with f = -0.936207316315, ‖∇f‖ = 7.6838e-05
[ Info: CG: iter    1: f = -0.936207316802, ‖∇f‖ = 3.3956e-05, α = 1.77e-01, β = 0.00e+00, nfg = 5
[ Info: CG: iter    2: f = -0.936207316979, ‖∇f‖ = 2.4905e-05, α = 3.25e-01, β = 1.65e-01, nfg = 2
[ Info: CG: iter    3: f = -0.936207317061, ‖∇f‖ = 1.3668e-05, α = 2.55e-01, β = 6.41e-01, nfg = 2
[ Info: CG: iter    4: f = -0.936207317085, ‖∇f‖ = 8.2456e-06, α = 2.51e-01, β = 3.10e-01, nfg = 2
[ Info: CG: iter    5: f = -0.936207317094, ‖∇f‖ = 5.4141e-06, α = 2.88e-01, β = 3.53e-01, nfg = 2
[ Info: CG: iter    6: f = -0.936207317098, ‖∇f‖ = 3.5886e-06, α = 2.58e-01, β = 4.29e-01, nfg = 2
[ Info: CG: iter    7: f = -0.936207317100, ‖∇f‖ = 2.3617e-06, α = 2.28e-01, β = 4.39e-01, nfg = 2
[ Info: CG: iter    8: f = -0.936207317100, ‖∇f‖ = 1.5488e-06, α = 2.07e-01, β = 4.38e-01, nfg = 2
[ Info: CG: iter    9: f = -0.936207317101, ‖∇f‖ = 1.0240e-06, α = 2.29e-01, β = 4.16e-01, nfg = 2
[ Info: CG: iter   10: f = -0.936207317101, ‖∇f‖ = 6.7229e-07, α = 2.44e-01, β = 4.52e-01, nfg = 2
[ Info: CG: iter   11: f = -0.936207317101, ‖∇f‖ = 4.5048e-07, α = 2.40e-01, β = 4.23e-01, nfg = 2
[ Info: CG: iter   12: f = -0.936207317101, ‖∇f‖ = 2.8542e-07, α = 2.10e-01, β = 4.53e-01, nfg = 2
[ Info: CG: iter   13: f = -0.936207317101, ‖∇f‖ = 1.7638e-07, α = 2.10e-01, β = 3.99e-01, nfg = 2
[ Info: CG: iter   14: f = -0.936207317101, ‖∇f‖ = 1.1830e-07, α = 2.33e-01, β = 3.84e-01, nfg = 2
[ Info: CG: iter   15: f = -0.936207317101, ‖∇f‖ = 8.0831e-08, α = 2.09e-01, β = 4.50e-01, nfg = 2
[ Info: CG: iter   16: f = -0.936207317101, ‖∇f‖ = 5.8991e-08, α = 2.06e-01, β = 4.67e-01, nfg = 2
[ Info: CG: iter   17: f = -0.936207317101, ‖∇f‖ = 4.3695e-08, α = 2.09e-01, β = 5.30e-01, nfg = 2
[ Info: CG: iter   18: f = -0.936207317101, ‖∇f‖ = 3.2093e-08, α = 1.97e-01, β = 5.52e-01, nfg = 2
[ Info: CG: iter   19: f = -0.936207317101, ‖∇f‖ = 2.3767e-08, α = 2.23e-01, β = 5.39e-01, nfg = 2
[ Info: CG: iter   20: f = -0.936207317101, ‖∇f‖ = 1.7986e-08, α = 2.39e-01, β = 5.49e-01, nfg = 2
[ Info: CG: iter   21: f = -0.936207317101, ‖∇f‖ = 1.3820e-08, α = 2.37e-01, β = 5.73e-01, nfg = 2
[ Info: CG: iter   22: f = -0.936207317101, ‖∇f‖ = 1.0492e-08, α = 2.20e-01, β = 5.91e-01, nfg = 2
[ Info: CG: iter   23: f = -0.936207317101, ‖∇f‖ = 8.5940e-09, α = 2.39e-01, β = 5.76e-01, nfg = 2
[ Info: CG: iter   24: f = -0.936207317101, ‖∇f‖ = 7.4331e-09, α = 2.67e-01, β = 6.72e-01, nfg = 2
[ Info: CG: iter   25: f = -0.936207317101, ‖∇f‖ = 6.0167e-09, α = 2.45e-01, β = 7.48e-01, nfg = 2
[ Info: CG: iter   26: f = -0.936207317101, ‖∇f‖ = 4.5352e-09, α = 2.12e-01, β = 6.55e-01, nfg = 2
[ Info: CG: iter   27: f = -0.936207317101, ‖∇f‖ = 2.9325e-09, α = 1.91e-01, β = 5.68e-01, nfg = 2
[ Info: CG: iter   28: f = -0.936207317101, ‖∇f‖ = 1.7500e-09, α = 1.76e-01, β = 4.18e-01, nfg = 2
[ Info: CG: iter   29: f = -0.936207317101, ‖∇f‖ = 1.0053e-09, α = 1.85e-01, β = 3.56e-01, nfg = 2
[ Info: CG: iter   30: f = -0.936207317101, ‖∇f‖ = 6.0296e-10, α = 2.02e-01, β = 3.30e-01, nfg = 2
[ Info: CG: iter   31: f = -0.936207317101, ‖∇f‖ = 3.9119e-10, α = 2.11e-01, β = 3.60e-01, nfg = 2
[ Info: CG: iter   32: f = -0.936207317101, ‖∇f‖ = 2.8042e-10, α = 2.19e-01, β = 4.21e-01, nfg = 2
[ Info: CG: iter   33: f = -0.936207317101, ‖∇f‖ = 2.1988e-10, α = 2.21e-01, β = 5.14e-01, nfg = 2
[ Info: CG: iter   34: f = -0.936207317101, ‖∇f‖ = 1.9308e-10, α = 2.56e-01, β = 6.15e-01, nfg = 2
[ Info: CG: iter   35: f = -0.936207317101, ‖∇f‖ = 1.6850e-10, α = 2.54e-01, β = 7.71e-01, nfg = 2
[ Info: CG: iter   36: f = -0.936207317101, ‖∇f‖ = 1.4140e-10, α = 2.25e-01, β = 7.61e-01, nfg = 2
[ Info: CG: iter   37: f = -0.936207317101, ‖∇f‖ = 1.1451e-10, α = 2.29e-01, β = 7.04e-01, nfg = 2
[ Info: CG: iter   38: f = -0.936207317101, ‖∇f‖ = 9.1346e-11, α = 2.33e-01, β = 6.55e-01, nfg = 2
[ Info: CG: iter   39: f = -0.936207317101, ‖∇f‖ = 7.0586e-11, α = 2.35e-01, β = 6.37e-01, nfg = 2
[ Info: CG: iter   40: f = -0.936207317101, ‖∇f‖ = 5.8120e-11, α = 2.60e-01, β = 5.97e-01, nfg = 2
[ Info: CG: iter   41: f = -0.936207317101, ‖∇f‖ = 4.7058e-11, α = 2.59e-01, β = 6.78e-01, nfg = 2
[ Info: CG: iter   42: f = -0.936207317101, ‖∇f‖ = 3.2408e-11, α = 2.24e-01, β = 6.55e-01, nfg = 2
[ Info: CG: iter   43: f = -0.936207317101, ‖∇f‖ = 2.1412e-11, α = 2.07e-01, β = 4.75e-01, nfg = 2
[ Info: CG: iter   44: f = -0.936207317101, ‖∇f‖ = 1.3598e-11, α = 1.86e-01, β = 4.35e-01, nfg = 2
[ Info: CG: iter   45: f = -0.936207317101, ‖∇f‖ = 8.9005e-12, α = 1.76e-01, β = 4.05e-01, nfg = 2
[ Info: CG: iter   46: f = -0.936207317101, ‖∇f‖ = 5.5988e-12, α = 1.85e-01, β = 4.27e-01, nfg = 2
[ Info: CG: iter   47: f = -0.936207317101, ‖∇f‖ = 3.6664e-12, α = 2.10e-01, β = 3.97e-01, nfg = 2
[ Info: CG: iter   48: f = -0.936207317101, ‖∇f‖ = 2.5881e-12, α = 2.24e-01, β = 4.29e-01, nfg = 2
[ Info: CG: iter   49: f = -0.936207317101, ‖∇f‖ = 1.9556e-12, α = 2.36e-01, β = 4.95e-01, nfg = 2
[ Info: CG: iter   50: f = -0.936207317101, ‖∇f‖ = 1.5149e-12, α = 2.25e-01, β = 5.73e-01, nfg = 2
[ Info: CG: iter   51: f = -0.936207317101, ‖∇f‖ = 1.2152e-12, α = 2.33e-01, β = 6.02e-01, nfg = 2
[ Info: CG: converged after 52 iterations: f = -0.936207317101, ‖∇f‖ = 9.9968e-13
[ Info: Found excitations for p = 2.9321531433504737
[ Info: Found excitations for p = 3.141592653589793
[ Info: Found excitations for p = 2.722713633111154
[ Info: Found excitations for p = 2.5132741228718345
[ Info: Found excitations for p = 2.303834612632515
[ Info: Found excitations for p = 0.0
[ Info: Found excitations for p = 2.0943951023931953
[ Info: Found excitations for p = 1.8849555921538759
[ Info: Found excitations for p = 1.6755160819145563
[ Info: Found excitations for p = 1.4660765716752369
[ Info: Found excitations for p = 0.8377580409572781
[ Info: Found excitations for p = 1.0471975511965976
[ Info: Found excitations for p = 1.2566370614359172
[ Info: Found excitations for p = 0.6283185307179586
[ Info: Found excitations for p = 0.41887902047863906
[ Info: Found excitations for p = 0.20943951023931953
minimum @k = 0.0:	 ΔE = -0.2595611997968351

````

````julia
plot(kspace, real.(Es); xaxis="momentum", yaxis="ΔE", label="S = 1")
````

```@raw html
<?xml version="1.0" encoding="utf-8"?>
<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" width="600" height="400" viewBox="0 0 2400 1600">
<defs>
  <clipPath id="clip400">
    <rect x="0" y="0" width="2400" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip400)" d="M0 1600 L2400 1600 L2400 0 L0 0  Z" fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip401">
    <rect x="480" y="0" width="1681" height="1600"/>
  </clipPath>
</defs>
<path clip-path="url(#clip400)" d="M102.74 1505.26 L2352.76 1505.26 L2352.76 47.2441 L102.74 47.2441  Z" fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<defs>
  <clipPath id="clip402">
    <rect x="102" y="47" width="2251" height="1459"/>
  </clipPath>
</defs>
<polyline clip-path="url(#clip402)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="166.42,1505.26 166.42,47.2441 "/>
<polyline clip-path="url(#clip402)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="842.082,1505.26 842.082,47.2441 "/>
<polyline clip-path="url(#clip402)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="1517.74,1505.26 1517.74,47.2441 "/>
<polyline clip-path="url(#clip402)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="2193.41,1505.26 2193.41,47.2441 "/>
<polyline clip-path="url(#clip400)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="102.74,1505.26 2352.76,1505.26 "/>
<polyline clip-path="url(#clip400)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="166.42,1505.26 166.42,1486.36 "/>
<polyline clip-path="url(#clip400)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="842.082,1505.26 842.082,1486.36 "/>
<polyline clip-path="url(#clip400)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="1517.74,1505.26 1517.74,1486.36 "/>
<polyline clip-path="url(#clip400)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="2193.41,1505.26 2193.41,1486.36 "/>
<path clip-path="url(#clip400)" d="M166.42 1536.18 Q162.809 1536.18 160.98 1539.74 Q159.175 1543.28 159.175 1550.41 Q159.175 1557.52 160.98 1561.09 Q162.809 1564.63 166.42 1564.63 Q170.054 1564.63 171.86 1561.09 Q173.688 1557.52 173.688 1550.41 Q173.688 1543.28 171.86 1539.74 Q170.054 1536.18 166.42 1536.18 M166.42 1532.47 Q172.23 1532.47 175.286 1537.08 Q178.364 1541.66 178.364 1550.41 Q178.364 1559.14 175.286 1563.75 Q172.23 1568.33 166.42 1568.33 Q160.61 1568.33 157.531 1563.75 Q154.475 1559.14 154.475 1550.41 Q154.475 1541.66 157.531 1537.08 Q160.61 1532.47 166.42 1532.47 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip400)" d="M832.464 1563.72 L840.103 1563.72 L840.103 1537.36 L831.793 1539.03 L831.793 1534.77 L840.057 1533.1 L844.733 1533.1 L844.733 1563.72 L852.372 1563.72 L852.372 1567.66 L832.464 1567.66 L832.464 1563.72 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip400)" d="M1512.4 1563.72 L1528.72 1563.72 L1528.72 1567.66 L1506.77 1567.66 L1506.77 1563.72 Q1509.43 1560.97 1514.02 1556.34 Q1518.62 1551.69 1519.81 1550.35 Q1522.05 1547.82 1522.93 1546.09 Q1523.83 1544.33 1523.83 1542.64 Q1523.83 1539.88 1521.89 1538.15 Q1519.97 1536.41 1516.87 1536.41 Q1514.67 1536.41 1512.21 1537.17 Q1509.78 1537.94 1507 1539.49 L1507 1534.77 Q1509.83 1533.63 1512.28 1533.05 Q1514.74 1532.47 1516.77 1532.47 Q1522.14 1532.47 1525.34 1535.16 Q1528.53 1537.85 1528.53 1542.34 Q1528.53 1544.47 1527.72 1546.39 Q1526.93 1548.28 1524.83 1550.88 Q1524.25 1551.55 1521.15 1554.77 Q1518.05 1557.96 1512.4 1563.72 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip400)" d="M2197.66 1549.03 Q2201.01 1549.74 2202.89 1552.01 Q2204.78 1554.28 2204.78 1557.61 Q2204.78 1562.73 2201.27 1565.53 Q2197.75 1568.33 2191.27 1568.33 Q2189.09 1568.33 2186.78 1567.89 Q2184.48 1567.47 2182.03 1566.62 L2182.03 1562.1 Q2183.97 1563.24 2186.29 1563.82 Q2188.6 1564.4 2191.13 1564.4 Q2195.53 1564.4 2197.82 1562.66 Q2200.13 1560.92 2200.13 1557.61 Q2200.13 1554.56 2197.98 1552.85 Q2195.85 1551.11 2192.03 1551.11 L2188 1551.11 L2188 1547.27 L2192.22 1547.27 Q2195.66 1547.27 2197.49 1545.9 Q2199.32 1544.51 2199.32 1541.92 Q2199.32 1539.26 2197.42 1537.85 Q2195.55 1536.41 2192.03 1536.41 Q2190.11 1536.41 2187.91 1536.83 Q2185.71 1537.24 2183.07 1538.12 L2183.07 1533.96 Q2185.73 1533.22 2188.05 1532.85 Q2190.39 1532.47 2192.45 1532.47 Q2197.77 1532.47 2200.87 1534.91 Q2203.97 1537.31 2203.97 1541.43 Q2203.97 1544.3 2202.33 1546.29 Q2200.69 1548.26 2197.66 1549.03 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip400)" d="M1072.12 1621.32 Q1074.32 1617.37 1077.37 1615.49 Q1080.43 1613.62 1084.57 1613.62 Q1090.14 1613.62 1093.16 1617.53 Q1096.18 1621.41 1096.18 1628.61 L1096.18 1650.12 L1090.3 1650.12 L1090.3 1628.8 Q1090.3 1623.67 1088.48 1621.19 Q1086.67 1618.71 1082.94 1618.71 Q1078.39 1618.71 1075.75 1621.73 Q1073.11 1624.76 1073.11 1629.98 L1073.11 1650.12 L1067.22 1650.12 L1067.22 1628.8 Q1067.22 1623.64 1065.41 1621.19 Q1063.59 1618.71 1059.8 1618.71 Q1055.32 1618.71 1052.68 1621.76 Q1050.03 1624.79 1050.03 1629.98 L1050.03 1650.12 L1044.15 1650.12 L1044.15 1614.48 L1050.03 1614.48 L1050.03 1620.01 Q1052.04 1616.74 1054.84 1615.18 Q1057.64 1613.62 1061.49 1613.62 Q1065.37 1613.62 1068.08 1615.59 Q1070.82 1617.56 1072.12 1621.32 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip400)" d="M1121.68 1618.58 Q1116.97 1618.58 1114.23 1622.27 Q1111.49 1625.93 1111.49 1632.33 Q1111.49 1638.73 1114.2 1642.42 Q1116.94 1646.08 1121.68 1646.08 Q1126.36 1646.08 1129.1 1642.39 Q1131.83 1638.7 1131.83 1632.33 Q1131.83 1626 1129.1 1622.31 Q1126.36 1618.58 1121.68 1618.58 M1121.68 1613.62 Q1129.32 1613.62 1133.68 1618.58 Q1138.04 1623.55 1138.04 1632.33 Q1138.04 1641.08 1133.68 1646.08 Q1129.32 1651.05 1121.68 1651.05 Q1114.01 1651.05 1109.65 1646.08 Q1105.32 1641.08 1105.32 1632.33 Q1105.32 1623.55 1109.65 1618.58 Q1114.01 1613.62 1121.68 1613.62 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip400)" d="M1175.5 1621.32 Q1177.7 1617.37 1180.75 1615.49 Q1183.81 1613.62 1187.95 1613.62 Q1193.52 1613.62 1196.54 1617.53 Q1199.56 1621.41 1199.56 1628.61 L1199.56 1650.12 L1193.68 1650.12 L1193.68 1628.8 Q1193.68 1623.67 1191.86 1621.19 Q1190.05 1618.71 1186.32 1618.71 Q1181.77 1618.71 1179.13 1621.73 Q1176.49 1624.76 1176.49 1629.98 L1176.49 1650.12 L1170.6 1650.12 L1170.6 1628.8 Q1170.6 1623.64 1168.79 1621.19 Q1166.97 1618.71 1163.18 1618.71 Q1158.7 1618.71 1156.05 1621.76 Q1153.41 1624.79 1153.41 1629.98 L1153.41 1650.12 L1147.52 1650.12 L1147.52 1614.48 L1153.41 1614.48 L1153.41 1620.01 Q1155.42 1616.74 1158.22 1615.18 Q1161.02 1613.62 1164.87 1613.62 Q1168.75 1613.62 1171.46 1615.59 Q1174.2 1617.56 1175.5 1621.32 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip400)" d="M1241.74 1630.84 L1241.74 1633.7 L1214.81 1633.7 Q1215.19 1639.75 1218.44 1642.93 Q1221.72 1646.08 1227.54 1646.08 Q1230.91 1646.08 1234.07 1645.25 Q1237.25 1644.43 1240.37 1642.77 L1240.37 1648.31 Q1237.22 1649.65 1233.91 1650.35 Q1230.6 1651.05 1227.19 1651.05 Q1218.66 1651.05 1213.66 1646.08 Q1208.7 1641.12 1208.7 1632.65 Q1208.7 1623.9 1213.41 1618.77 Q1218.15 1613.62 1226.17 1613.62 Q1233.37 1613.62 1237.54 1618.26 Q1241.74 1622.88 1241.74 1630.84 M1235.88 1629.12 Q1235.82 1624.31 1233.17 1621.45 Q1230.56 1618.58 1226.24 1618.58 Q1221.33 1618.58 1218.37 1621.35 Q1215.45 1624.12 1215 1629.15 L1235.88 1629.12 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip400)" d="M1280.98 1628.61 L1280.98 1650.12 L1275.12 1650.12 L1275.12 1628.8 Q1275.12 1623.74 1273.15 1621.22 Q1271.18 1618.71 1267.23 1618.71 Q1262.49 1618.71 1259.75 1621.73 Q1257.01 1624.76 1257.01 1629.98 L1257.01 1650.12 L1251.13 1650.12 L1251.13 1614.48 L1257.01 1614.48 L1257.01 1620.01 Q1259.12 1616.8 1261.95 1615.21 Q1264.81 1613.62 1268.54 1613.62 Q1274.68 1613.62 1277.83 1617.44 Q1280.98 1621.22 1280.98 1628.61 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip400)" d="M1298.46 1604.35 L1298.46 1614.48 L1310.52 1614.48 L1310.52 1619.03 L1298.46 1619.03 L1298.46 1638.38 Q1298.46 1642.74 1299.63 1643.98 Q1300.84 1645.22 1304.5 1645.22 L1310.52 1645.22 L1310.52 1650.12 L1304.5 1650.12 Q1297.72 1650.12 1295.14 1647.61 Q1292.57 1645.06 1292.57 1638.38 L1292.57 1619.03 L1288.27 1619.03 L1288.27 1614.48 L1292.57 1614.48 L1292.57 1604.35 L1298.46 1604.35 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip400)" d="M1317.62 1636.06 L1317.62 1614.48 L1323.47 1614.48 L1323.47 1635.83 Q1323.47 1640.89 1325.45 1643.44 Q1327.42 1645.95 1331.37 1645.95 Q1336.11 1645.95 1338.85 1642.93 Q1341.61 1639.91 1341.61 1634.69 L1341.61 1614.48 L1347.47 1614.48 L1347.47 1650.12 L1341.61 1650.12 L1341.61 1644.65 Q1339.48 1647.9 1336.65 1649.49 Q1333.85 1651.05 1330.12 1651.05 Q1323.98 1651.05 1320.8 1647.23 Q1317.62 1643.41 1317.62 1636.06 M1332.35 1613.62 L1332.35 1613.62 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip400)" d="M1387.29 1621.32 Q1389.48 1617.37 1392.54 1615.49 Q1395.6 1613.62 1399.73 1613.62 Q1405.3 1613.62 1408.33 1617.53 Q1411.35 1621.41 1411.35 1628.61 L1411.35 1650.12 L1405.46 1650.12 L1405.46 1628.8 Q1405.46 1623.67 1403.65 1621.19 Q1401.83 1618.71 1398.11 1618.71 Q1393.56 1618.71 1390.92 1621.73 Q1388.28 1624.76 1388.28 1629.98 L1388.28 1650.12 L1382.39 1650.12 L1382.39 1628.8 Q1382.39 1623.64 1380.57 1621.19 Q1378.76 1618.71 1374.97 1618.71 Q1370.48 1618.71 1367.84 1621.76 Q1365.2 1624.79 1365.2 1629.98 L1365.2 1650.12 L1359.31 1650.12 L1359.31 1614.48 L1365.2 1614.48 L1365.2 1620.01 Q1367.2 1616.74 1370.01 1615.18 Q1372.81 1613.62 1376.66 1613.62 Q1380.54 1613.62 1383.25 1615.59 Q1385.98 1617.56 1387.29 1621.32 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><polyline clip-path="url(#clip402)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="102.74,1377.41 2352.76,1377.41 "/>
<polyline clip-path="url(#clip402)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="102.74,1086.67 2352.76,1086.67 "/>
<polyline clip-path="url(#clip402)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="102.74,795.927 2352.76,795.927 "/>
<polyline clip-path="url(#clip402)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="102.74,505.185 2352.76,505.185 "/>
<polyline clip-path="url(#clip402)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:2; stroke-opacity:0.1; fill:none" points="102.74,214.443 2352.76,214.443 "/>
<polyline clip-path="url(#clip400)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="102.74,1505.26 102.74,47.2441 "/>
<polyline clip-path="url(#clip400)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="102.74,1377.41 121.638,1377.41 "/>
<polyline clip-path="url(#clip400)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="102.74,1086.67 121.638,1086.67 "/>
<polyline clip-path="url(#clip400)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="102.74,795.927 121.638,795.927 "/>
<polyline clip-path="url(#clip400)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="102.74,505.185 121.638,505.185 "/>
<polyline clip-path="url(#clip400)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="102.74,214.443 121.638,214.443 "/>
<path clip-path="url(#clip400)" d="M-38.6065 1377.86 L-8.93073 1377.86 L-8.93073 1381.8 L-38.6065 1381.8 L-38.6065 1377.86 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip400)" d="M11.1618 1363.21 Q7.55066 1363.21 5.72197 1366.77 Q3.91642 1370.32 3.91642 1377.45 Q3.91642 1384.55 5.72197 1388.12 Q7.55066 1391.66 11.1618 1391.66 Q14.796 1391.66 16.6015 1388.12 Q18.4302 1384.55 18.4302 1377.45 Q18.4302 1370.32 16.6015 1366.77 Q14.796 1363.21 11.1618 1363.21 M11.1618 1359.51 Q16.9719 1359.51 20.0274 1364.11 Q23.1061 1368.7 23.1061 1377.45 Q23.1061 1386.17 20.0274 1390.78 Q16.9719 1395.36 11.1618 1395.36 Q5.3516 1395.36 2.27291 1390.78 Q-0.782629 1386.17 -0.782629 1377.45 Q-0.782629 1368.7 2.27291 1364.11 Q5.3516 1359.51 11.1618 1359.51 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip400)" d="M31.3237 1388.81 L36.2079 1388.81 L36.2079 1394.69 L31.3237 1394.69 L31.3237 1388.81 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip400)" d="M50.4208 1390.76 L66.7402 1390.76 L66.7402 1394.69 L44.7958 1394.69 L44.7958 1390.76 Q47.4579 1388 52.0412 1383.37 Q56.6476 1378.72 57.8282 1377.38 Q60.0735 1374.85 60.9532 1373.12 Q61.8559 1371.36 61.8559 1369.67 Q61.8559 1366.91 59.9115 1365.18 Q57.9902 1363.44 54.8884 1363.44 Q52.6893 1363.44 50.2356 1364.2 Q47.8051 1364.97 45.0273 1366.52 L45.0273 1361.8 Q47.8514 1360.66 50.3051 1360.08 Q52.7588 1359.51 54.7958 1359.51 Q60.1661 1359.51 63.3605 1362.19 Q66.555 1364.88 66.555 1369.37 Q66.555 1371.5 65.7448 1373.42 Q64.9578 1375.32 62.8513 1377.91 Q62.2726 1378.58 59.1708 1381.8 Q56.0689 1384.99 50.4208 1390.76 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip400)" d="M9.56454 1072.47 Q5.95345 1072.47 4.12475 1076.03 Q2.31921 1079.57 2.31921 1086.7 Q2.31921 1093.81 4.12475 1097.37 Q5.95345 1100.92 9.56454 1100.92 Q13.1988 1100.92 15.0043 1097.37 Q16.833 1093.81 16.833 1086.7 Q16.833 1079.57 15.0043 1076.03 Q13.1988 1072.47 9.56454 1072.47 M9.56454 1068.76 Q15.3747 1068.76 18.4302 1073.37 Q21.5089 1077.95 21.5089 1086.7 Q21.5089 1095.43 18.4302 1100.04 Q15.3747 1104.62 9.56454 1104.62 Q3.75438 1104.62 0.675697 1100.04 Q-2.37984 1095.43 -2.37984 1086.7 Q-2.37984 1077.95 0.675697 1073.37 Q3.75438 1068.76 9.56454 1068.76 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip400)" d="M29.7265 1098.07 L34.6107 1098.07 L34.6107 1103.95 L29.7265 1103.95 L29.7265 1098.07 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip400)" d="M54.7958 1072.47 Q51.1847 1072.47 49.356 1076.03 Q47.5504 1079.57 47.5504 1086.7 Q47.5504 1093.81 49.356 1097.37 Q51.1847 1100.92 54.7958 1100.92 Q58.43 1100.92 60.2356 1097.37 Q62.0643 1093.81 62.0643 1086.7 Q62.0643 1079.57 60.2356 1076.03 Q58.43 1072.47 54.7958 1072.47 M54.7958 1068.76 Q60.6059 1068.76 63.6615 1073.37 Q66.7402 1077.95 66.7402 1086.7 Q66.7402 1095.43 63.6615 1100.04 Q60.6059 1104.62 54.7958 1104.62 Q48.9856 1104.62 45.9069 1100.04 Q42.8514 1095.43 42.8514 1086.7 Q42.8514 1077.95 45.9069 1073.37 Q48.9856 1068.76 54.7958 1068.76 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip400)" d="M11.1618 781.725 Q7.55066 781.725 5.72197 785.29 Q3.91642 788.832 3.91642 795.961 Q3.91642 803.068 5.72197 806.633 Q7.55066 810.174 11.1618 810.174 Q14.796 810.174 16.6015 806.633 Q18.4302 803.068 18.4302 795.961 Q18.4302 788.832 16.6015 785.29 Q14.796 781.725 11.1618 781.725 M11.1618 778.022 Q16.9719 778.022 20.0274 782.628 Q23.1061 787.211 23.1061 795.961 Q23.1061 804.688 20.0274 809.295 Q16.9719 813.878 11.1618 813.878 Q5.3516 813.878 2.27291 809.295 Q-0.782629 804.688 -0.782629 795.961 Q-0.782629 787.211 2.27291 782.628 Q5.3516 778.022 11.1618 778.022 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip400)" d="M31.3237 807.327 L36.2079 807.327 L36.2079 813.207 L31.3237 813.207 L31.3237 807.327 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip400)" d="M50.4208 809.272 L66.7402 809.272 L66.7402 813.207 L44.7958 813.207 L44.7958 809.272 Q47.4579 806.517 52.0412 801.887 Q56.6476 797.235 57.8282 795.892 Q60.0735 793.369 60.9532 791.633 Q61.8559 789.873 61.8559 788.184 Q61.8559 785.429 59.9115 783.693 Q57.9902 781.957 54.8884 781.957 Q52.6893 781.957 50.2356 782.721 Q47.8051 783.485 45.0273 785.036 L45.0273 780.313 Q47.8514 779.179 50.3051 778.6 Q52.7588 778.022 54.7958 778.022 Q60.1661 778.022 63.3605 780.707 Q66.555 783.392 66.555 787.883 Q66.555 790.012 65.7448 791.934 Q64.9578 793.832 62.8513 796.424 Q62.2726 797.096 59.1708 800.313 Q56.0689 803.508 50.4208 809.272 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip400)" d="M9.07843 490.983 Q5.46734 490.983 3.63864 494.548 Q1.8331 498.09 1.8331 505.219 Q1.8331 512.326 3.63864 515.891 Q5.46734 519.432 9.07843 519.432 Q12.7127 519.432 14.5182 515.891 Q16.3469 512.326 16.3469 505.219 Q16.3469 498.09 14.5182 494.548 Q12.7127 490.983 9.07843 490.983 M9.07843 487.28 Q14.8886 487.28 17.9441 491.886 Q21.0228 496.47 21.0228 505.219 Q21.0228 513.946 17.9441 518.553 Q14.8886 523.136 9.07843 523.136 Q3.26828 523.136 0.189588 518.553 Q-2.86595 513.946 -2.86595 505.219 Q-2.86595 496.47 0.189588 491.886 Q3.26828 487.28 9.07843 487.28 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip400)" d="M29.2404 516.585 L34.1246 516.585 L34.1246 522.465 L29.2404 522.465 L29.2404 516.585 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip400)" d="M57.1569 491.979 L45.3514 510.428 L57.1569 510.428 L57.1569 491.979 M55.93 487.905 L61.8096 487.905 L61.8096 510.428 L66.7402 510.428 L66.7402 514.317 L61.8096 514.317 L61.8096 522.465 L57.1569 522.465 L57.1569 514.317 L41.5551 514.317 L41.5551 509.803 L55.93 487.905 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip400)" d="M9.4025 200.242 Q5.79141 200.242 3.96272 203.806 Q2.15717 207.348 2.15717 214.478 Q2.15717 221.584 3.96272 225.149 Q5.79141 228.69 9.4025 228.69 Q13.0367 228.69 14.8423 225.149 Q16.671 221.584 16.671 214.478 Q16.671 207.348 14.8423 203.806 Q13.0367 200.242 9.4025 200.242 M9.4025 196.538 Q15.2127 196.538 18.2682 201.144 Q21.3469 205.728 21.3469 214.478 Q21.3469 223.204 18.2682 227.811 Q15.2127 232.394 9.4025 232.394 Q3.59235 232.394 0.51366 227.811 Q-2.54188 223.204 -2.54188 214.478 Q-2.54188 205.728 0.51366 201.144 Q3.59235 196.538 9.4025 196.538 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip400)" d="M29.5644 225.843 L34.4487 225.843 L34.4487 231.723 L29.5644 231.723 L29.5644 225.843 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip400)" d="M55.2124 212.579 Q52.0643 212.579 50.2125 214.732 Q48.3838 216.885 48.3838 220.635 Q48.3838 224.362 50.2125 226.538 Q52.0643 228.69 55.2124 228.69 Q58.3606 228.69 60.1893 226.538 Q62.0411 224.362 62.0411 220.635 Q62.0411 216.885 60.1893 214.732 Q58.3606 212.579 55.2124 212.579 M64.4948 197.927 L64.4948 202.186 Q62.7355 201.353 60.93 200.913 Q59.1476 200.473 57.3884 200.473 Q52.7588 200.473 50.3051 203.598 Q47.8745 206.723 47.5273 213.042 Q48.893 211.028 50.9532 209.964 Q53.0134 208.876 55.4902 208.876 Q60.6985 208.876 63.7078 212.047 Q66.7402 215.195 66.7402 220.635 Q66.7402 225.959 63.592 229.177 Q60.4439 232.394 55.2124 232.394 Q49.2171 232.394 46.0458 227.811 Q42.8745 223.204 42.8745 214.478 Q42.8745 206.283 46.7634 201.422 Q50.6523 196.538 57.2032 196.538 Q58.9624 196.538 60.7448 196.885 Q62.5504 197.232 64.4948 197.927 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip400)" d="M-130.049 795.031 L-94.2097 808.081 L-94.2097 781.949 L-130.049 795.031 M-88.8625 816.802 L-136.383 798.659 L-136.383 791.371 L-88.8625 773.26 L-88.8625 816.802 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip400)" d="M-136.383 766.321 L-136.383 736.275 L-130.972 736.275 L-130.972 759.892 L-116.903 759.892 L-116.903 737.262 L-111.493 737.262 L-111.493 759.892 L-94.2734 759.892 L-94.2734 735.702 L-88.8625 735.702 L-88.8625 766.321 L-136.383 766.321 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><polyline clip-path="url(#clip402)" style="stroke:#009af9; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="166.42,1464 307.93,1068.71 449.441,985.025 590.951,838.035 732.462,849.507 873.972,677.099 1015.48,537.618 1156.99,482.762 1298.5,372.519 1440.01,289.781 1581.52,240.497 1723.03,182.407 1864.54,138.926 2006.06,115.645 2147.57,97.2925 2289.08,88.5087 "/>
<path clip-path="url(#clip400)" d="M177.741 199.525 L525.357 199.525 L525.357 95.8446 L177.741 95.8446  Z" fill="#ffffff" fill-rule="evenodd" fill-opacity="1"/>
<polyline clip-path="url(#clip400)" style="stroke:#000000; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="177.741,199.525 525.357,199.525 525.357,95.8446 177.741,95.8446 177.741,199.525 "/>
<polyline clip-path="url(#clip400)" style="stroke:#009af9; stroke-linecap:round; stroke-linejoin:round; stroke-width:4; stroke-opacity:1; fill:none" points="202.741,147.685 352.742,147.685 "/>
<path clip-path="url(#clip400)" d="M399.987 131.539 L399.987 136.099 Q397.325 134.826 394.964 134.201 Q392.603 133.576 390.404 133.576 Q386.585 133.576 384.501 135.057 Q382.441 136.539 382.441 139.27 Q382.441 141.562 383.807 142.743 Q385.196 143.9 389.038 144.618 L391.862 145.196 Q397.094 146.192 399.571 148.715 Q402.071 151.215 402.071 155.428 Q402.071 160.451 398.691 163.043 Q395.335 165.636 388.83 165.636 Q386.376 165.636 383.599 165.08 Q380.844 164.525 377.881 163.437 L377.881 158.622 Q380.728 160.219 383.46 161.029 Q386.191 161.84 388.83 161.84 Q392.835 161.84 395.011 160.266 Q397.186 158.692 397.186 155.775 Q397.186 153.229 395.612 151.793 Q394.061 150.358 390.497 149.641 L387.649 149.085 Q382.418 148.043 380.08 145.821 Q377.742 143.599 377.742 139.641 Q377.742 135.057 380.96 132.418 Q384.2 129.78 389.872 129.78 Q392.302 129.78 394.825 130.219 Q397.348 130.659 399.987 131.539 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip400)" d="M424.802 143.437 L454.478 143.437 L454.478 147.326 L424.802 147.326 L424.802 143.437 M424.802 152.881 L454.478 152.881 L454.478 156.817 L424.802 156.817 L424.802 152.881 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /><path clip-path="url(#clip400)" d="M480.45 161.029 L488.089 161.029 L488.089 134.664 L479.779 136.331 L479.779 132.071 L488.042 130.405 L492.718 130.405 L492.718 161.029 L500.357 161.029 L500.357 164.965 L480.45 164.965 L480.45 161.029 Z" fill="#000000" fill-rule="nonzero" fill-opacity="1" /></svg>

```

---

*This page was generated using [Literate.jl](https://github.com/fredrikekre/Literate.jl).*

